@@ -569,27 +569,14 @@ impl super::Instruction {
         instruction
     }
 
-    pub(super) fn image_fetch(
+    pub(super) fn image_fetch_or_read(
+        op: Op,
         result_type_id: Word,
         id: Word,
         image: Word,
         coordinates: Word,
     ) -> Self {
-        let mut instruction = Self::new(Op::ImageFetch);
-        instruction.set_type(result_type_id);
-        instruction.set_result(id);
-        instruction.add_operand(image);
-        instruction.add_operand(coordinates);
-        instruction
-    }
-
-    pub(super) fn image_read(
-        result_type_id: Word,
-        id: Word,
-        image: Word,
-        coordinates: Word,
-    ) -> Self {
-        let mut instruction = Self::new(Op::ImageRead);
+        let mut instruction = Self::new(op);
         instruction.set_type(result_type_id);
         instruction.set_result(id);
         instruction.add_operand(image);
@@ -716,6 +703,42 @@ impl super::Instruction {
         instruction
     }
 
+    pub(super) fn ternary(
+        op: Op,
+        result_type_id: Word,
+        id: Word,
+        operand_1: Word,
+        operand_2: Word,
+        operand_3: Word,
+    ) -> Self {
+        let mut instruction = Self::new(op);
+        instruction.set_type(result_type_id);
+        instruction.set_result(id);
+        instruction.add_operand(operand_1);
+        instruction.add_operand(operand_2);
+        instruction.add_operand(operand_3);
+        instruction
+    }
+
+    pub(super) fn quaternary(
+        op: Op,
+        result_type_id: Word,
+        id: Word,
+        operand_1: Word,
+        operand_2: Word,
+        operand_3: Word,
+        operand_4: Word,
+    ) -> Self {
+        let mut instruction = Self::new(op);
+        instruction.set_type(result_type_id);
+        instruction.set_result(id);
+        instruction.add_operand(operand_1);
+        instruction.add_operand(operand_2);
+        instruction.add_operand(operand_3);
+        instruction.add_operand(operand_4);
+        instruction
+    }
+
     pub(super) fn relational(op: Op, result_type_id: Word, id: Word, expr_id: Word) -> Self {
         let mut instruction = Self::new(op);
         instruction.set_type(result_type_id);
