@@ -3,16 +3,21 @@ use super::{
     make_local, Block, BlockContext, CachedExpressions, EntryPointContext, Error, Function,
     FunctionArgument, GlobalVariable, IdGenerator, Instruction, LocalType, LocalVariable,
     LogicalLayout, LookupFunctionType, LookupType, LoopContext, Options, PhysicalLayout,
-    ResultMember, Writer, WriterFlags, BITS_PER_BYTE,
+    PipelineOptions, ResultMember, Writer, WriterFlags, BITS_PER_BYTE,
 };
 use crate::{
-    arena::{Arena, Handle},
+    arena::{Handle, UniqueArena},
     proc::TypeResolution,
     valid::{FunctionInfo, ModuleInfo},
 };
 use spirv::Word;
 use std::collections::hash_map::Entry;
 
+struct FunctionInterface<'a> {
+    varying_ids: &'a mut Vec<Word>,
+    stage: crate::ShaderStage,
+}
+
 impl Function {
     fn to_words(&self, sink: &mut impl Extend<Word>) {
         self.signature.as_ref().unwrap().to_words(sink);
@@ -57,7 +62,7 @@ impl Writer {
             debugs: vec![],
             annotations: vec![],
             flags: options.flags,
-            index_bounds_check_policy: options.index_bounds_check_policy,
+            bounds_check_policies: options.bounds_check_policies,
             void_type,
             lookup_type: crate::FastHashMap::default(),
             lookup_function: crate::FastHashMap::default(),
@@ -93,7 +98,7 @@ impl Writer {
         let fresh = Writer {
             // Copied from the old Writer:
             flags: self.flags,
-            index_bounds_check_policy: self.index_bounds_check_policy,
+            bounds_check_policies: self.bounds_check_policies,
             capabilities_available: take(&mut self.capabilities_available),
 
             // Initialized afresh:
@@ -190,7 +195,7 @@ impl Writer {
 
     pub(super) fn get_pointer_id(
         &mut self,
-        arena: &Arena<crate::Type>,
+        arena: &UniqueArena<crate::Type>,
         handle: Handle<crate::Type>,
         class: spirv::StorageClass,
     ) -> Result<Word, Error> {
@@ -223,6 +228,35 @@ impl Writer {
         self.get_type_id(local_type.into())
     }
 
+    pub(super) fn get_float_type_id(&mut self) -> Word {
+        let local_type = LocalType::Value {
+            vector_size: None,
+            kind: crate::ScalarKind::Float,
+            width: 4,
+            pointer_class: None,
+        };
+        self.get_type_id(local_type.into())
+    }
+
+    pub(super) fn get_float_pointer_type_id(&mut self, class: spirv::StorageClass) -> Word {
+        let lookup_type = LookupType::Local(LocalType::Value {
+            vector_size: None,
+            kind: crate::ScalarKind::Float,
+            width: 4,
+            pointer_class: Some(class),
+        });
+        if let Some(&id) = self.lookup_type.get(&lookup_type) {
+            id
+        } else {
+            let id = self.id_gen.next();
+            let ty_id = self.get_float_type_id();
+            let instruction = Instruction::type_pointer(id, class, ty_id);
+            instruction.to_words(&mut self.logical_layout.declarations);
+            self.lookup_type.insert(lookup_type, id);
+            id
+        }
+    }
+
     pub(super) fn get_bool_type_id(&mut self) -> Word {
         let local_type = LocalType::Value {
             vector_size: None,
@@ -243,7 +277,7 @@ impl Writer {
         ir_function: &crate::Function,
         info: &FunctionInfo,
         ir_module: &crate::Module,
-        mut varying_ids: Option<&mut Vec<Word>>,
+        mut interface: Option<FunctionInterface>,
     ) -> Result<Word, Error> {
         let mut function = Function::default();
 
@@ -291,12 +325,13 @@ impl Writer {
                 )?,
                 false => self.get_type_id(LookupType::Handle(argument.ty)),
             };
-            if let Some(ref mut list) = varying_ids {
+
+            if let Some(ref mut iface) = interface {
                 let id = if let Some(ref binding) = argument.binding {
                     let name = argument.name.as_ref().map(AsRef::as_ref);
                     let varying_id =
                         self.write_varying(ir_module, class, name, argument.ty, binding)?;
-                    list.push(varying_id);
+                    iface.varying_ids.push(varying_id);
                     let id = self.id_gen.next();
                     prelude
                         .body
@@ -313,7 +348,7 @@ impl Writer {
                         let binding = member.binding.as_ref().unwrap();
                         let varying_id =
                             self.write_varying(ir_module, class, name, member.ty, binding)?;
-                        list.push(varying_id);
+                        iface.varying_ids.push(varying_id);
                         let id = self.id_gen.next();
                         prelude
                             .body
@@ -354,13 +389,16 @@ impl Writer {
 
         let return_type_id = match ir_function.result {
             Some(ref result) => {
-                if let Some(ref mut list) = varying_ids {
+                if let Some(ref mut iface) = interface {
+                    let mut has_point_size = false;
                     let class = spirv::StorageClass::Output;
                     if let Some(ref binding) = result.binding {
+                        has_point_size |=
+                            *binding == crate::Binding::BuiltIn(crate::BuiltIn::PointSize);
                         let type_id = self.get_type_id(LookupType::Handle(result.ty));
                         let varying_id =
                             self.write_varying(ir_module, class, None, result.ty, binding)?;
-                        list.push(varying_id);
+                        iface.varying_ids.push(varying_id);
                         ep_context.results.push(ResultMember {
                             id: varying_id,
                             type_id,
@@ -373,9 +411,11 @@ impl Writer {
                             let type_id = self.get_type_id(LookupType::Handle(member.ty));
                             let name = member.name.as_ref().map(AsRef::as_ref);
                             let binding = member.binding.as_ref().unwrap();
+                            has_point_size |=
+                                *binding == crate::Binding::BuiltIn(crate::BuiltIn::PointSize);
                             let varying_id =
                                 self.write_varying(ir_module, class, name, member.ty, binding)?;
-                            list.push(varying_id);
+                            iface.varying_ids.push(varying_id);
                             ep_context.results.push(ResultMember {
                                 id: varying_id,
                                 type_id,
@@ -385,6 +425,29 @@ impl Writer {
                     } else {
                         unreachable!("Missing result binding on an entry point");
                     }
+
+                    if self.flags.contains(WriterFlags::FORCE_POINT_SIZE)
+                        && iface.stage == crate::ShaderStage::Vertex
+                        && !has_point_size
+                    {
+                        // add point size artificially
+                        let varying_id = self.id_gen.next();
+                        let pointer_type_id = self.get_float_pointer_type_id(class);
+                        Instruction::variable(pointer_type_id, varying_id, class, None)
+                            .to_words(&mut self.logical_layout.declarations);
+                        self.decorate(
+                            varying_id,
+                            spirv::Decoration::BuiltIn,
+                            &[spirv::BuiltIn::PointSize as u32],
+                        );
+                        iface.varying_ids.push(varying_id);
+
+                        let default_value_id =
+                            self.get_constant_scalar(crate::ScalarValue::Float(1.0), 4);
+                        prelude
+                            .body
+                            .push(Instruction::store(varying_id, default_value_id, None));
+                    }
                     self.void_type
                 } else {
                     self.get_type_id(LookupType::Handle(result.ty))
@@ -413,13 +476,13 @@ impl Writer {
             function_type,
         ));
 
-        if varying_ids.is_some() {
+        if interface.is_some() {
             function.entry_point_context = Some(ep_context);
         }
 
         // fill up the `GlobalVariable::handle_id`
         for gv in self.global_variables.iter_mut() {
-            gv.handle_id = 0;
+            gv.reset_for_function();
         }
         for (handle, var) in ir_module.global_variables.iter() {
             // Handle globals are pre-emitted and should be loaded automatically.
@@ -501,7 +564,10 @@ impl Writer {
             &entry_point.function,
             info,
             ir_module,
-            Some(&mut interface_ids),
+            Some(FunctionInterface {
+                varying_ids: &mut interface_ids,
+                stage: entry_point.stage,
+            }),
         )?;
 
         let exec_model = match entry_point.stage {
@@ -700,7 +766,7 @@ impl Writer {
 
     fn write_type_declaration_arena(
         &mut self,
-        arena: &Arena<crate::Type>,
+        arena: &UniqueArena<crate::Type>,
         handle: Handle<crate::Type>,
     ) -> Result<Word, Error> {
         let ty = &arena[handle];
@@ -893,7 +959,7 @@ impl Writer {
         }
     }
 
-    pub(super) fn get_index_constant(&mut self, index: Word) -> Result<Word, Error> {
+    pub(super) fn get_index_constant(&mut self, index: Word) -> Word {
         self.get_constant_scalar(crate::ScalarValue::Uint(index as _), 4)
     }
 
@@ -901,14 +967,14 @@ impl Writer {
         &mut self,
         value: crate::ScalarValue,
         width: crate::Bytes,
-    ) -> Result<Word, Error> {
+    ) -> Word {
         if let Some(&id) = self.cached_constants.get(&(value, width)) {
-            return Ok(id);
+            return id;
         }
         let id = self.id_gen.next();
-        self.write_constant_scalar(id, &value, width, None)?;
+        self.write_constant_scalar(id, &value, width, None);
         self.cached_constants.insert((value, width), id);
-        Ok(id)
+        id
     }
 
     fn write_constant_scalar(
@@ -917,7 +983,7 @@ impl Writer {
         value: &crate::ScalarValue,
         width: crate::Bytes,
         debug_name: Option<&String>,
-    ) -> Result<(), Error> {
+    ) {
         if self.flags.contains(WriterFlags::DEBUG) {
             if let Some(name) = debug_name {
                 self.debugs.push(Instruction::name(id, name));
@@ -979,7 +1045,6 @@ impl Writer {
         };
 
         instruction.to_words(&mut self.logical_layout.declarations);
-        Ok(())
     }
 
     fn write_constant_composite(
@@ -1094,6 +1159,10 @@ impl Writer {
                             BuiltIn::FragCoord
                         }
                     }
+                    Bi::ViewIndex => {
+                        self.require_any("`view_index` built-in", &[spirv::Capability::MultiView])?;
+                        BuiltIn::ViewIndex
+                    }
                     // vertex
                     Bi::BaseInstance => BuiltIn::BaseInstance,
                     Bi::BaseVertex => BuiltIn::BaseVertex,
@@ -1217,7 +1286,21 @@ impl Writer {
         &mut self,
         ir_module: &crate::Module,
         mod_info: &ModuleInfo,
+        ep_index: Option<usize>,
     ) -> Result<(), Error> {
+        fn has_view_index_check(
+            ir_module: &crate::Module,
+            binding: Option<&crate::Binding>,
+            ty: Handle<crate::Type>,
+        ) -> bool {
+            match ir_module.types[ty].inner {
+                crate::TypeInner::Struct { ref members, .. } => members.iter().any(|member| {
+                    has_view_index_check(ir_module, member.binding.as_ref(), member.ty)
+                }),
+                _ => binding == Some(&crate::Binding::BuiltIn(crate::BuiltIn::ViewIndex)),
+            }
+        }
+
         let has_storage_buffers =
             ir_module
                 .global_variables
@@ -1226,11 +1309,21 @@ impl Writer {
                     crate::StorageClass::Storage { .. } => true,
                     _ => false,
                 });
+        let has_view_index = ir_module
+            .entry_points
+            .iter()
+            .flat_map(|entry| entry.function.arguments.iter())
+            .any(|arg| has_view_index_check(ir_module, arg.binding.as_ref(), arg.ty));
+
         if self.physical_layout.version < 0x10300 && has_storage_buffers {
             // enable the storage buffer class on < SPV-1.3
             Instruction::extension("SPV_KHR_storage_buffer_storage_class")
                 .to_words(&mut self.logical_layout.extensions);
         }
+        if has_view_index {
+            Instruction::extension("SPV_KHR_multiview")
+                .to_words(&mut self.logical_layout.extensions)
+        }
         Instruction::type_void(self.void_type).to_words(&mut self.logical_layout.declarations);
         Instruction::ext_inst_import(self.gl450_ext_inst_id, "GLSL.std.450")
             .to_words(&mut self.logical_layout.ext_inst_imports);
@@ -1249,10 +1342,10 @@ impl Writer {
                     self.constant_ids[handle.index()] = match constant.name {
                         Some(ref name) => {
                             let id = self.id_gen.next();
-                            self.write_constant_scalar(id, value, width, Some(name))?;
+                            self.write_constant_scalar(id, value, width, Some(name));
                             id
                         }
-                        None => self.get_constant_scalar(*value, width)?,
+                        None => self.get_constant_scalar(*value, width),
                     };
                 }
             }
@@ -1282,23 +1375,46 @@ impl Writer {
         debug_assert_eq!(self.constant_ids.iter().position(|&id| id == 0), None);
 
         // now write all globals
-        for (_, var) in ir_module.global_variables.iter() {
-            let (instruction, id) = self.write_global_variable(ir_module, var)?;
-            instruction.to_words(&mut self.logical_layout.declarations);
-            self.global_variables
-                .push(GlobalVariable { id, handle_id: 0 });
+        for (handle, var) in ir_module.global_variables.iter() {
+            // If a single entry point was specified, only write `OpVariable` instructions
+            // for the globals it actually uses. Emit dummies for the others,
+            // to preserve the indices in `global_variables`.
+            let gvar = match ep_index {
+                Some(index) if mod_info.get_entry_point(index)[handle].is_empty() => {
+                    GlobalVariable::dummy()
+                }
+                _ => {
+                    let (instruction, id) = self.write_global_variable(ir_module, var)?;
+                    instruction.to_words(&mut self.logical_layout.declarations);
+                    GlobalVariable::new(id)
+                }
+            };
+            self.global_variables.push(gvar);
         }
 
         // all functions
         for (handle, ir_function) in ir_module.functions.iter() {
             let info = &mod_info[handle];
+            if let Some(index) = ep_index {
+                let ep_info = mod_info.get_entry_point(index);
+                // If this function uses globals that we omitted from the SPIR-V
+                // because the entry point and its callees didn't use them,
+                // then we must skip it.
+                if !ep_info.dominates_global_use(info) {
+                    log::info!("Skip function {:?}", ir_function.name);
+                    continue;
+                }
+            }
             let id = self.write_function(ir_function, info, ir_module, None)?;
             self.lookup_function.insert(handle, id);
         }
 
         // and entry points
-        for (ep_index, ir_ep) in ir_module.entry_points.iter().enumerate() {
-            let info = mod_info.get_entry_point(ep_index);
+        for (index, ir_ep) in ir_module.entry_points.iter().enumerate() {
+            if ep_index.is_some() && ep_index != Some(index) {
+                continue;
+            }
+            let info = mod_info.get_entry_point(index);
             let ep_instruction = self.write_entry_point(ir_ep, info, ir_module)?;
             ep_instruction.to_words(&mut self.logical_layout.entry_points);
         }
@@ -1337,11 +1453,25 @@ impl Writer {
         &mut self,
         ir_module: &crate::Module,
         info: &ModuleInfo,
+        pipeline_options: Option<&PipelineOptions>,
         words: &mut Vec<Word>,
     ) -> Result<(), Error> {
         self.reset();
 
-        self.write_logical_layout(ir_module, info)?;
+        // Try to find the entry point and corresponding index
+        let ep_index = match pipeline_options {
+            Some(po) => {
+                let index = ir_module
+                    .entry_points
+                    .iter()
+                    .position(|ep| po.shader_stage == ep.stage && po.entry_point == ep.name)
+                    .ok_or(Error::EntryPointNotFound)?;
+                Some(index)
+            }
+            None => None,
+        };
+
+        self.write_logical_layout(ir_module, info, ep_index)?;
         self.write_physical_layout();
 
         self.physical_layout.in_words(words);
