@@ -22,6 +22,35 @@ enum Attribute {
     WorkGroupSize([u32; 3]),
 }
 
+/// The WGSL form that `write_expr_with_indirection` should use to render a Naga
+/// expression.
+///
+/// Sometimes a Naga `Expression` alone doesn't provide enough information to
+/// choose the right rendering for it in WGSL. For example, one natural WGSL
+/// rendering of a Naga `LocalVariable(x)` expression might be `&x`, since
+/// `LocalVariable` produces a pointer to the local variable's storage. But when
+/// rendering a `Store` statement, the `pointer` operand must be the left hand
+/// side of a WGSL assignment, so the proper rendering is `x`.
+///
+/// The caller of `write_expr_with_indirection` must provide an `Expected` value
+/// to indicate how ambiguous expressions should be rendered.
+#[derive(Clone, Copy, Debug)]
+enum Indirection {
+    /// Render pointer-construction expressions as WGSL `ptr`-typed expressions.
+    ///
+    /// This is the right choice for most cases. Whenever a Naga pointer
+    /// expression is not the `pointer` operand of a `Load` or `Store`, it
+    /// must be a WGSL pointer expression.
+    Ordinary,
+
+    /// Render pointer-construction expressions as WGSL reference-typed
+    /// expressions.
+    ///
+    /// For example, this is the right choice for the `pointer` operand when
+    /// rendering a `Store` statement as a WGSL assignment.
+    Reference,
+}
+
 pub struct Writer<W> {
     out: W,
     names: crate::FastHashMap<NameKey, String>,
@@ -496,10 +525,13 @@ impl<W: Write> Writer<W> {
                         "storage_",
                         "",
                         storage_format_str(format),
-                        if access.contains(crate::StorageAccess::STORE) {
-                            ",write"
+                        if access.contains(crate::StorageAccess::LOAD | crate::StorageAccess::STORE)
+                        {
+                            ",read_write"
+                        } else if access.contains(crate::StorageAccess::LOAD) {
+                            ",read"
                         } else {
-                            ""
+                            ",write"
                         },
                     ),
                 };
@@ -551,17 +583,67 @@ impl<W: Write> Writer<W> {
             }
             TypeInner::Pointer { base, class } => {
                 let (storage, maybe_access) = storage_class_str(class);
+                // Everything but `StorageClass::Handle` gives us a `storage` name, but
+                // Naga IR never produces pointers to handles, so it doesn't matter much
+                // how we write such a type. Just write it as the base type alone.
                 if let Some(class) = storage {
                     write!(self.out, "ptr<{}, ", class)?;
-                    if let Some(access) = maybe_access {
-                        write!(self.out, ", {}", access)?;
-                    }
                 }
                 self.write_type(module, base)?;
                 if storage.is_some() {
+                    if let Some(access) = maybe_access {
+                        write!(self.out, ", {}", access)?;
+                    }
                     write!(self.out, ">")?;
                 }
             }
+            TypeInner::ValuePointer {
+                size: None,
+                kind,
+                width: _,
+                class,
+            } => {
+                let (storage, maybe_access) = storage_class_str(class);
+                if let Some(class) = storage {
+                    write!(self.out, "ptr<{}, {}", class, scalar_kind_str(kind))?;
+                    if let Some(access) = maybe_access {
+                        write!(self.out, ", {}", access)?;
+                    }
+                    write!(self.out, ">")?;
+                } else {
+                    return Err(Error::Unimplemented(format!(
+                        "ValuePointer to StorageClass::Handle {:?}",
+                        inner
+                    )));
+                }
+            }
+            TypeInner::ValuePointer {
+                size: Some(size),
+                kind,
+                width: _,
+                class,
+            } => {
+                let (storage, maybe_access) = storage_class_str(class);
+                if let Some(class) = storage {
+                    write!(
+                        self.out,
+                        "ptr<{}, vec{}<{}>",
+                        class,
+                        back::vector_size_str(size),
+                        scalar_kind_str(kind)
+                    )?;
+                    if let Some(access) = maybe_access {
+                        write!(self.out, ", {}", access)?;
+                    }
+                    write!(self.out, ">")?;
+                } else {
+                    return Err(Error::Unimplemented(format!(
+                        "ValuePointer to StorageClass::Handle {:?}",
+                        inner
+                    )));
+                }
+                write!(self.out, ">")?;
+            }
             _ => {
                 return Err(Error::Unimplemented(format!(
                     "write_value_type {:?}",
@@ -588,12 +670,13 @@ impl<W: Write> Writer<W> {
         match *stmt {
             Statement::Emit(ref range) => {
                 for handle in range.clone() {
+                    let info = &func_ctx.info[handle];
                     let expr_name = if let Some(name) = func_ctx.named_expressions.get(&handle) {
                         // Front end provides names for all variables at the start of writing.
                         // But we write them to step by step. We need to recache them
                         // Otherwise, we could accidentally write variable name instead of full expression.
                         // Also, we use sanitized names! It defense backend from generating variable with name from reserved keywords.
-                        Some(self.namer.call_unique(name))
+                        Some(self.namer.call(name))
                     } else {
                         let expr = &func_ctx.expressions[handle];
                         let min_ref_count = expr.bake_ref_count();
@@ -604,8 +687,7 @@ impl<W: Write> Writer<W> {
                             | Expression::ImageSample { .. } => true,
                             _ => false,
                         };
-                        if min_ref_count <= func_ctx.info[handle].ref_count || required_baking_expr
-                        {
+                        if min_ref_count <= info.ref_count || required_baking_expr {
                             // If expression contains unsupported builtin we should skip it
                             if let Expression::Load { pointer } = func_ctx.expressions[handle] {
                                 if let Expression::AccessIndex { base, index } =
@@ -622,7 +704,7 @@ impl<W: Write> Writer<W> {
                                 }
                             }
 
-                            Some(format!("{}{}", back::BAKE_PREFIX, handle.index()))
+                            Some(format!("{}{}", super::BAKE_PREFIX, handle.index()))
                         } else {
                             None
                         }
@@ -693,28 +775,26 @@ impl<W: Write> Writer<W> {
                 }
                 write!(self.out, "{}", level)?;
 
-                let (is_ptr, is_atomic) = match *func_ctx.info[pointer].ty.inner_with(&module.types)
-                {
-                    crate::TypeInner::Pointer { base, .. } => (
-                        func_ctx.expressions[pointer].should_deref(),
-                        match module.types[base].inner {
-                            crate::TypeInner::Atomic { .. } => true,
-                            _ => false,
-                        },
-                    ),
-                    _ => (false, false),
+                let is_atomic = match *func_ctx.info[pointer].ty.inner_with(&module.types) {
+                    crate::TypeInner::Pointer { base, .. } => match module.types[base].inner {
+                        crate::TypeInner::Atomic { .. } => true,
+                        _ => false,
+                    },
+                    _ => false,
                 };
                 if is_atomic {
                     write!(self.out, "atomicStore(")?;
-                    if !is_ptr {
-                        write!(self.out, "&")?;
-                    }
                     self.write_expr(module, pointer, func_ctx)?;
                     write!(self.out, ", ")?;
                     self.write_expr(module, value, func_ctx)?;
                     write!(self.out, ")")?;
                 } else {
-                    self.write_expr(module, pointer, func_ctx)?;
+                    self.write_expr_with_indirection(
+                        module,
+                        pointer,
+                        func_ctx,
+                        Indirection::Reference,
+                    )?;
                     write!(self.out, " = ")?;
                     self.write_expr(module, value, func_ctx)?;
                 }
@@ -727,14 +807,14 @@ impl<W: Write> Writer<W> {
             } => {
                 write!(self.out, "{}", level)?;
                 if let Some(expr) = result {
-                    let name = format!("{}{}", back::BAKE_PREFIX, expr.index());
+                    let name = format!("{}{}", super::BAKE_PREFIX, expr.index());
                     self.start_named_expr(module, expr, func_ctx, &name)?;
                     self.named_expressions.insert(expr, name);
                 }
                 let func_name = &self.names[&NameKey::Function(function)];
                 write!(self.out, "{}(", func_name)?;
-                for (index, argument) in arguments.iter().enumerate() {
-                    self.write_expr(module, *argument, func_ctx)?;
+                for (index, &argument) in arguments.iter().enumerate() {
+                    self.write_expr(module, argument, func_ctx)?;
                     // Only write a comma if isn't the last element
                     if index != arguments.len().saturating_sub(1) {
                         // The leading space is for readability only
@@ -750,17 +830,12 @@ impl<W: Write> Writer<W> {
                 result,
             } => {
                 write!(self.out, "{}", level)?;
-                let res_name = format!("{}{}", back::BAKE_PREFIX, result.index());
+                let res_name = format!("{}{}", super::BAKE_PREFIX, result.index());
                 self.start_named_expr(module, result, func_ctx, &res_name)?;
                 self.named_expressions.insert(result, res_name);
 
-                let is_ptr = func_ctx.expressions[pointer].should_deref();
-
                 let fun_str = fun.to_wgsl();
                 write!(self.out, "atomic{}(", fun_str)?;
-                if !is_ptr {
-                    write!(self.out, "&")?;
-                }
                 self.write_expr(module, pointer, func_ctx)?;
                 if let crate::AtomicFunction::Exchange { compare: Some(cmp) } = *fun {
                     write!(self.out, ", ")?;
@@ -815,6 +890,13 @@ impl<W: Write> Writer<W> {
                 let all_fall_through = cases
                     .iter()
                     .all(|case| case.fall_through && case.body.is_empty());
+                let type_postfix = match *func_ctx.info[selector].ty.inner_with(&module.types) {
+                    crate::TypeInner::Scalar {
+                        kind: crate::ScalarKind::Uint,
+                        ..
+                    } => "u",
+                    _ => "",
+                };
 
                 let l2 = level.next();
                 if !cases.is_empty() {
@@ -824,11 +906,11 @@ impl<W: Write> Writer<W> {
                         }
                         if !all_fall_through && case.fall_through && case.body.is_empty() {
                             write_case = false;
-                            write!(self.out, "{}, ", case.value)?;
+                            write!(self.out, "{}{}, ", case.value, type_postfix)?;
                             continue;
                         } else {
                             write_case = true;
-                            writeln!(self.out, "{}: {{", case.value)?;
+                            writeln!(self.out, "{}{}: {{", case.value, type_postfix)?;
                         }
 
                         for sta in case.body.iter() {
@@ -897,6 +979,65 @@ impl<W: Write> Writer<W> {
         Ok(())
     }
 
+    /// Return the sort of indirection that `expr`'s plain form evaluates to.
+    ///
+    /// An expression's 'plain form' is the most general rendition of that
+    /// expression into WGSL, lacking `&` or `*` operators:
+    ///
+    /// - The plain form of `LocalVariable(x)` is simply `x`, which is a reference
+    ///   to the local variable's storage.
+    ///
+    /// - The plain form of `GlobalVariable(g)` is simply `g`, which is usually a
+    ///   reference to the global variable's storage. However, globals in the
+    ///   `Handle` storage class are immutable, and `GlobalVariable` expressions for
+    ///   those produce the value directly, not a pointer to it. Such
+    ///   `GlobalVariable` expressions are `Ordinary`.
+    ///
+    /// - `Access` and `AccessIndex` are `Reference` when their `base` operand is a
+    ///   pointer. If they are applied directly to a composite value, they are
+    ///   `Ordinary`.
+    ///
+    /// Note that `FunctionArgument` expressions are never `Reference`, even when
+    /// the argument's type is `Pointer`. `FunctionArgument` always evaluates to the
+    /// argument's value directly, so any pointer it produces is merely the value
+    /// passed by the caller.
+    fn plain_form_indirection(
+        &self,
+        expr: Handle<crate::Expression>,
+        module: &Module,
+        func_ctx: &back::FunctionCtx<'_>,
+    ) -> Indirection {
+        use crate::Expression as Ex;
+
+        // Named expressions are `let` expressions, which apply the Load Rule,
+        // so if their type is a Naga pointer, then that must be a WGSL pointer
+        // as well.
+        if self.named_expressions.contains_key(&expr) {
+            return Indirection::Ordinary;
+        }
+
+        match func_ctx.expressions[expr] {
+            Ex::LocalVariable(_) => Indirection::Reference,
+            Ex::GlobalVariable(handle) => {
+                let global = &module.global_variables[handle];
+                match global.class {
+                    crate::StorageClass::Handle => Indirection::Ordinary,
+                    _ => Indirection::Reference,
+                }
+            }
+            Ex::Access { base, .. } | Ex::AccessIndex { base, .. } => {
+                let base_ty = func_ctx.info[base].ty.inner_with(&module.types);
+                match *base_ty {
+                    crate::TypeInner::Pointer { .. } | crate::TypeInner::ValuePointer { .. } => {
+                        Indirection::Reference
+                    }
+                    _ => Indirection::Ordinary,
+                }
+            }
+            _ => Indirection::Ordinary,
+        }
+    }
+
     fn start_named_expr(
         &mut self,
         module: &Module,
@@ -921,15 +1062,70 @@ impl<W: Write> Writer<W> {
         Ok(())
     }
 
-    /// Helper method to write expressions
+    /// Write the ordinary WGSL form of `expr`.
     ///
-    /// # Notes
-    /// Doesn't add any newlines or leading/trailing spaces
+    /// See `write_expr_with_indirection` for details.
     fn write_expr(
         &mut self,
         module: &Module,
         expr: Handle<crate::Expression>,
         func_ctx: &back::FunctionCtx<'_>,
+    ) -> BackendResult {
+        self.write_expr_with_indirection(module, expr, func_ctx, Indirection::Ordinary)
+    }
+
+    /// Write `expr` as a WGSL expression with the requested indirection.
+    ///
+    /// In terms of the WGSL grammar, the resulting expression is a
+    /// `singular_expression`. It may be parenthesized. This makes it suitable
+    /// for use as the operand of a unary or binary operator without worrying
+    /// about precedence.
+    ///
+    /// This does not produce newlines or indentation.
+    ///
+    /// The `requested` argument indicates (roughly) whether Naga
+    /// `Pointer`-valued expressions represent WGSL references or pointers. See
+    /// `Indirection` for details.
+    fn write_expr_with_indirection(
+        &mut self,
+        module: &Module,
+        expr: Handle<crate::Expression>,
+        func_ctx: &back::FunctionCtx<'_>,
+        requested: Indirection,
+    ) -> BackendResult {
+        // If the plain form of the expression is not what we need, emit the
+        // operator necessary to correct that.
+        let plain = self.plain_form_indirection(expr, module, func_ctx);
+        match (requested, plain) {
+            (Indirection::Ordinary, Indirection::Reference) => {
+                write!(self.out, "(&")?;
+                self.write_expr_plain_form(module, expr, func_ctx, plain)?;
+                write!(self.out, ")")?;
+            }
+            (Indirection::Reference, Indirection::Ordinary) => {
+                write!(self.out, "(*")?;
+                self.write_expr_plain_form(module, expr, func_ctx, plain)?;
+                write!(self.out, ")")?;
+            }
+            (_, _) => self.write_expr_plain_form(module, expr, func_ctx, plain)?,
+        }
+
+        Ok(())
+    }
+
+    /// Write the 'plain form' of `expr`.
+    ///
+    /// An expression's 'plain form' is the most general rendition of that
+    /// expression into WGSL, lacking `&` or `*` operators. The plain forms of
+    /// `LocalVariable(x)` and `GlobalVariable(g)` are simply `x` and `g`. Such
+    /// Naga expressions represent both WGSL pointers and references; it's the
+    /// caller's responsibility to distinguish those cases appropriately.
+    fn write_expr_plain_form(
+        &mut self,
+        module: &Module,
+        expr: Handle<crate::Expression>,
+        func_ctx: &back::FunctionCtx<'_>,
+        indirection: Indirection,
     ) -> BackendResult {
         use crate::Expression;
 
@@ -940,6 +1136,14 @@ impl<W: Write> Writer<W> {
 
         let expression = &func_ctx.expressions[expr];
 
+        // Write the plain WGSL form of a Naga expression.
+        //
+        // The plain form of `LocalVariable` and `GlobalVariable` expressions is
+        // simply the variable name; `*` and `&` operators are never emitted.
+        //
+        // The plain form of `Access` and `AccessIndex` expressions are WGSL
+        // `postfix_expression` forms for member/component access and
+        // subscripting.
         match *expression {
             Expression::Constant(constant) => self.write_constant(module, constant)?,
             Expression::Compose { ty, ref components } => {
@@ -999,30 +1203,17 @@ impl<W: Write> Writer<W> {
                 self.write_expr(module, right, func_ctx)?;
                 write!(self.out, ")")?;
             }
-            // TODO: copy-paste from glsl-out
             Expression::Access { base, index } => {
-                self.write_expr(module, base, func_ctx)?;
+                self.write_expr_with_indirection(module, base, func_ctx, indirection)?;
                 write!(self.out, "[")?;
                 self.write_expr(module, index, func_ctx)?;
                 write!(self.out, "]")?
             }
-            // TODO: copy-paste from glsl-out
             Expression::AccessIndex { base, index } => {
                 let base_ty_res = &func_ctx.info[base].ty;
                 let mut resolved = base_ty_res.inner_with(&module.types);
 
-                let deref = match *resolved {
-                    TypeInner::Pointer { .. } => func_ctx.expressions[base].should_deref(),
-                    _ => false,
-                };
-
-                if deref {
-                    write!(self.out, "(*")?;
-                }
-                self.write_expr(module, base, func_ctx)?;
-                if deref {
-                    write!(self.out, ")")?;
-                }
+                self.write_expr_with_indirection(module, base, func_ctx, indirection)?;
 
                 let base_ty_handle = match *resolved {
                     TypeInner::Pointer { base, class: _ } => {
@@ -1166,7 +1357,11 @@ impl<W: Write> Writer<W> {
                 let name = &self.names[&NameKey::GlobalVariable(handle)];
                 write!(self.out, "{}", name)?;
             }
-            Expression::As { expr, kind, .. } => {
+            Expression::As {
+                expr,
+                kind,
+                convert,
+            } => {
                 let inner = func_ctx.info[expr].ty.inner_with(&module.types);
                 match *inner {
                     TypeInner::Matrix { columns, rows, .. } => {
@@ -1178,14 +1373,25 @@ impl<W: Write> Writer<W> {
                         )?;
                     }
                     TypeInner::Vector { size, .. } => {
-                        write!(
-                            self.out,
-                            "vec{}<{}>",
-                            back::vector_size_str(size),
-                            scalar_kind_str(kind)
-                        )?;
+                        let vector_size_str = back::vector_size_str(size);
+                        let scalar_kind_str = scalar_kind_str(kind);
+                        if convert.is_some() {
+                            write!(self.out, "vec{}<{}>", vector_size_str, scalar_kind_str)?;
+                        } else {
+                            write!(
+                                self.out,
+                                "bitcast<vec{}<{}>>",
+                                vector_size_str, scalar_kind_str
+                            )?;
+                        }
+                    }
+                    TypeInner::Scalar { .. } => {
+                        if convert.is_some() {
+                            write!(self.out, "{}", scalar_kind_str(kind))?
+                        } else {
+                            write!(self.out, "bitcast<{}>", scalar_kind_str(kind))?
+                        }
                     }
-                    TypeInner::Scalar { .. } => write!(self.out, "{}", scalar_kind_str(kind))?,
                     _ => {
                         return Err(Error::Unimplemented(format!(
                             "write_expr expression::as {:?}",
@@ -1216,30 +1422,25 @@ impl<W: Write> Writer<W> {
                 write!(self.out, ")")?;
             }
             Expression::Load { pointer } => {
-                let (is_pointer, is_atomic) =
-                    match *func_ctx.info[pointer].ty.inner_with(&module.types) {
-                        crate::TypeInner::Pointer { base, .. } => (
-                            func_ctx.expressions[pointer].should_deref(),
-                            match module.types[base].inner {
-                                crate::TypeInner::Atomic { .. } => true,
-                                _ => false,
-                            },
-                        ),
-                        _ => (false, false),
-                    };
+                let is_atomic = match *func_ctx.info[pointer].ty.inner_with(&module.types) {
+                    crate::TypeInner::Pointer { base, .. } => match module.types[base].inner {
+                        crate::TypeInner::Atomic { .. } => true,
+                        _ => false,
+                    },
+                    _ => false,
+                };
+
                 if is_atomic {
                     write!(self.out, "atomicLoad(")?;
-                    if !is_pointer {
-                        // Write an indirection in case the underlying
-                        // expression isn't a pointer but a reference
-                        write!(self.out, "&")?;
-                    }
-                } else if is_pointer {
-                    write!(self.out, "*")?;
-                }
-                self.write_expr(module, pointer, func_ctx)?;
-                if is_atomic {
+                    self.write_expr(module, pointer, func_ctx)?;
                     write!(self.out, ")")?;
+                } else {
+                    self.write_expr_with_indirection(
+                        module,
+                        pointer,
+                        func_ctx,
+                        Indirection::Reference,
+                    )?;
                 }
             }
             Expression::LocalVariable(handle) => {
@@ -1247,9 +1448,6 @@ impl<W: Write> Writer<W> {
             }
             Expression::ArrayLength(expr) => {
                 write!(self.out, "arrayLength(")?;
-                if is_deref_required(expr, module, func_ctx.info) {
-                    write!(self.out, "&")?;
-                };
                 self.write_expr(module, expr, func_ctx)?;
                 write!(self.out, ")")?;
             }
@@ -1258,78 +1456,131 @@ impl<W: Write> Writer<W> {
                 arg,
                 arg1,
                 arg2,
+                arg3,
             } => {
                 use crate::MathFunction as Mf;
 
-                let fun_name = match fun {
-                    Mf::Abs => "abs",
-                    Mf::Min => "min",
-                    Mf::Max => "max",
-                    Mf::Clamp => "clamp",
+                enum Function {
+                    Asincosh { is_sin: bool },
+                    Atanh,
+                    Regular(&'static str),
+                }
+
+                // NOTE: If https://github.com/gpuweb/gpuweb/issues/1622 ever is
+                // accepted, replace this with the builtin functions
+                let function = match fun {
+                    Mf::Abs => Function::Regular("abs"),
+                    Mf::Min => Function::Regular("min"),
+                    Mf::Max => Function::Regular("max"),
+                    Mf::Clamp => Function::Regular("clamp"),
                     // trigonometry
-                    Mf::Cos => "cos",
-                    Mf::Cosh => "cosh",
-                    Mf::Sin => "sin",
-                    Mf::Sinh => "sinh",
-                    Mf::Tan => "tan",
-                    Mf::Tanh => "tanh",
-                    Mf::Acos => "acos",
-                    Mf::Asin => "asin",
-                    Mf::Atan => "atan",
-                    Mf::Atan2 => "atan2",
+                    Mf::Cos => Function::Regular("cos"),
+                    Mf::Cosh => Function::Regular("cosh"),
+                    Mf::Sin => Function::Regular("sin"),
+                    Mf::Sinh => Function::Regular("sinh"),
+                    Mf::Tan => Function::Regular("tan"),
+                    Mf::Tanh => Function::Regular("tanh"),
+                    Mf::Acos => Function::Regular("acos"),
+                    Mf::Asin => Function::Regular("asin"),
+                    Mf::Atan => Function::Regular("atan"),
+                    Mf::Atan2 => Function::Regular("atan2"),
+                    Mf::Asinh => Function::Asincosh { is_sin: true },
+                    Mf::Acosh => Function::Asincosh { is_sin: false },
+                    Mf::Atanh => Function::Atanh,
                     // decomposition
-                    Mf::Ceil => "ceil",
-                    Mf::Floor => "floor",
-                    Mf::Round => "round",
-                    Mf::Fract => "fract",
-                    Mf::Trunc => "trunc",
-                    Mf::Modf => "modf",
-                    Mf::Frexp => "frexp",
-                    Mf::Ldexp => "ldexp",
+                    Mf::Ceil => Function::Regular("ceil"),
+                    Mf::Floor => Function::Regular("floor"),
+                    Mf::Round => Function::Regular("round"),
+                    Mf::Fract => Function::Regular("fract"),
+                    Mf::Trunc => Function::Regular("trunc"),
+                    Mf::Modf => Function::Regular("modf"),
+                    Mf::Frexp => Function::Regular("frexp"),
+                    Mf::Ldexp => Function::Regular("ldexp"),
                     // exponent
-                    Mf::Exp => "exp",
-                    Mf::Exp2 => "exp2",
-                    Mf::Log => "log",
-                    Mf::Log2 => "log2",
-                    Mf::Pow => "pow",
+                    Mf::Exp => Function::Regular("exp"),
+                    Mf::Exp2 => Function::Regular("exp2"),
+                    Mf::Log => Function::Regular("log"),
+                    Mf::Log2 => Function::Regular("log2"),
+                    Mf::Pow => Function::Regular("pow"),
                     // geometry
-                    Mf::Dot => "dot",
-                    Mf::Outer => "outerProduct",
-                    Mf::Cross => "cross",
-                    Mf::Distance => "distance",
-                    Mf::Length => "length",
-                    Mf::Normalize => "normalize",
-                    Mf::FaceForward => "faceForward",
-                    Mf::Reflect => "reflect",
+                    Mf::Dot => Function::Regular("dot"),
+                    Mf::Outer => Function::Regular("outerProduct"),
+                    Mf::Cross => Function::Regular("cross"),
+                    Mf::Distance => Function::Regular("distance"),
+                    Mf::Length => Function::Regular("length"),
+                    Mf::Normalize => Function::Regular("normalize"),
+                    Mf::FaceForward => Function::Regular("faceForward"),
+                    Mf::Reflect => Function::Regular("reflect"),
                     // computational
-                    Mf::Sign => "sign",
-                    Mf::Fma => "fma",
-                    Mf::Mix => "mix",
-                    Mf::Step => "step",
-                    Mf::SmoothStep => "smoothStep",
-                    Mf::Sqrt => "sqrt",
-                    Mf::InverseSqrt => "inverseSqrt",
-                    Mf::Transpose => "transpose",
-                    Mf::Determinant => "determinant",
+                    Mf::Sign => Function::Regular("sign"),
+                    Mf::Fma => Function::Regular("fma"),
+                    Mf::Mix => Function::Regular("mix"),
+                    Mf::Step => Function::Regular("step"),
+                    Mf::SmoothStep => Function::Regular("smoothStep"),
+                    Mf::Sqrt => Function::Regular("sqrt"),
+                    Mf::InverseSqrt => Function::Regular("inverseSqrt"),
+                    Mf::Transpose => Function::Regular("transpose"),
+                    Mf::Determinant => Function::Regular("determinant"),
                     // bits
-                    Mf::CountOneBits => "countOneBits",
-                    Mf::ReverseBits => "reverseBits",
+                    Mf::CountOneBits => Function::Regular("countOneBits"),
+                    Mf::ReverseBits => Function::Regular("reverseBits"),
+                    Mf::ExtractBits => Function::Regular("extractBits"),
+                    Mf::InsertBits => Function::Regular("insertBits"),
+                    // data packing
+                    Mf::Pack4x8snorm => Function::Regular("pack4x8snorm"),
+                    Mf::Pack4x8unorm => Function::Regular("pack4x8unorm"),
+                    Mf::Pack2x16snorm => Function::Regular("pack2x16snorm"),
+                    Mf::Pack2x16unorm => Function::Regular("pack2x16unorm"),
+                    Mf::Pack2x16float => Function::Regular("pack2x16float"),
+                    // data unpacking
+                    Mf::Unpack4x8snorm => Function::Regular("unpack4x8snorm"),
+                    Mf::Unpack4x8unorm => Function::Regular("unpack4x8unorm"),
+                    Mf::Unpack2x16snorm => Function::Regular("unpack2x16snorm"),
+                    Mf::Unpack2x16unorm => Function::Regular("unpack2x16unorm"),
+                    Mf::Unpack2x16float => Function::Regular("unpack2x16float"),
                     _ => {
                         return Err(Error::UnsupportedMathFunction(fun));
                     }
                 };
 
-                write!(self.out, "{}(", fun_name)?;
-                self.write_expr(module, arg, func_ctx)?;
-                if let Some(arg) = arg1 {
-                    write!(self.out, ", ")?;
-                    self.write_expr(module, arg, func_ctx)?;
+                match function {
+                    Function::Asincosh { is_sin } => {
+                        write!(self.out, "log(")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, " + sqrt(")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, " * ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        match is_sin {
+                            true => write!(self.out, " + 1.0))")?,
+                            false => write!(self.out, " - 1.0))")?,
+                        }
+                    }
+                    Function::Atanh => {
+                        write!(self.out, "0.5 * log((1.0 + ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, ") / (1.0 - ")?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        write!(self.out, "))")?;
+                    }
+                    Function::Regular(fun_name) => {
+                        write!(self.out, "{}(", fun_name)?;
+                        self.write_expr(module, arg, func_ctx)?;
+                        if let Some(arg) = arg1 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        if let Some(arg) = arg2 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        if let Some(arg) = arg3 {
+                            write!(self.out, ", ")?;
+                            self.write_expr(module, arg, func_ctx)?;
+                        }
+                        write!(self.out, ")")?
+                    }
                 }
-                if let Some(arg) = arg2 {
-                    write!(self.out, ", ")?;
-                    self.write_expr(module, arg, func_ctx)?;
-                }
-                write!(self.out, ")")?
             }
             Expression::Swizzle {
                 size,
@@ -1588,21 +1839,6 @@ impl<W: Write> Writer<W> {
     }
 }
 
-impl crate::Expression {
-    /// Wether an expression should be dereferenced, this is false when the
-    /// expression returns a reference instead of a pointer
-    fn should_deref(&self) -> bool {
-        match *self {
-            // Variables in the typifier have pointer types but in wgsl they
-            // have reference types and shouldn't be dereferenced
-            crate::Expression::LocalVariable(_) | crate::Expression::GlobalVariable(_)
-            // Access chains might have pointer types but wgsl considers them as references
-            | crate::Expression::AccessIndex {..} | crate::Expression::Access {..} => false,
-            _ => true,
-        }
-    }
-}
-
 fn builtin_str(built_in: crate::BuiltIn) -> Option<&'static str> {
     use crate::BuiltIn as Bi;
 
@@ -1753,23 +1989,6 @@ fn map_binding_to_attribute(
     }
 }
 
-fn is_deref_required(
-    expr: Handle<crate::Expression>,
-    module: &Module,
-    info: &valid::FunctionInfo,
-) -> bool {
-    let base_ty_res = &info[expr].ty;
-    let resolved = base_ty_res.inner_with(&module.types);
-    match *resolved {
-        TypeInner::Pointer { base, class: _ } => match module.types[base].inner {
-            TypeInner::Scalar { .. } | TypeInner::Vector { .. } | TypeInner::Array { .. } => true,
-            _ => false,
-        },
-        TypeInner::ValuePointer { .. } => true,
-        _ => false,
-    }
-}
-
 /// Helper function that check that expression don't access to structure member with unsupported builtin.
 fn access_to_unsupported_builtin(
     expr: Handle<crate::Expression>,
