@@ -1,12 +1,12 @@
 use crate::{
-    arena::{Arena, Handle},
+    arena::{Arena, Handle, UniqueArena},
     BinaryOperator, Constant, ConstantInner, Expression, ScalarKind, ScalarValue, Type, TypeInner,
     UnaryOperator,
 };
 
 #[derive(Debug)]
 pub struct ConstantSolver<'a> {
-    pub types: &'a Arena<Type>,
+    pub types: &'a mut UniqueArena<Type>,
     pub expressions: &'a Arena<Expression>,
     pub constants: &'a mut Arena<Constant>,
 }
@@ -53,8 +53,10 @@ pub enum ConstantSolvingError {
     InvalidBinaryOpArgs,
     #[error("Cannot apply math function to type")]
     InvalidMathArg,
-    #[error("Splat/swizzle type is not registered")]
-    DestinationTypeNotFound,
+    #[error("Splat is defined only on scalar values")]
+    SplatScalarOnly,
+    #[error("Can only swizzle vector constants")]
+    SwizzleVectorOnly,
     #[error("Not implemented: {0}")]
     NotImplemented(String),
 }
@@ -64,7 +66,7 @@ impl<'a> ConstantSolver<'a> {
         &mut self,
         expr: Handle<Expression>,
     ) -> Result<Handle<Constant>, ConstantSolvingError> {
-        let span = self.expressions.get_span(expr).clone();
+        let span = self.expressions.get_span(expr);
         match self.expressions[expr] {
             Expression::Constant(constant) => Ok(constant),
             Expression::AccessIndex { base, index } => self.access(base, index as usize),
@@ -81,14 +83,19 @@ impl<'a> ConstantSolver<'a> {
                 let ty = match self.constants[value_constant].inner {
                     ConstantInner::Scalar { ref value, width } => {
                         let kind = value.scalar_kind();
-                        self.types
-                            .fetch_if(|t| t.inner == crate::TypeInner::Vector { size, kind, width })
+                        self.types.insert(
+                            Type {
+                                name: None,
+                                inner: TypeInner::Vector { size, kind, width },
+                            },
+                            span,
+                        )
+                    }
+                    ConstantInner::Composite { .. } => {
+                        return Err(ConstantSolvingError::SplatScalarOnly);
                     }
-                    ConstantInner::Composite { .. } => None,
                 };
 
-                //TODO: register the new type if needed
-                let ty = ty.ok_or(ConstantSolvingError::DestinationTypeNotFound)?;
                 Ok(self.constants.fetch_or_append(
                     Constant {
                         name: None,
@@ -108,7 +115,9 @@ impl<'a> ConstantSolver<'a> {
             } => {
                 let src_constant = self.solve(src_vector)?;
                 let (ty, src_components) = match self.constants[src_constant].inner {
-                    ConstantInner::Scalar { .. } => (None, &[][..]),
+                    ConstantInner::Scalar { .. } => {
+                        return Err(ConstantSolvingError::SwizzleVectorOnly);
+                    }
                     ConstantInner::Composite {
                         ty,
                         components: ref src_components,
@@ -118,16 +127,21 @@ impl<'a> ConstantSolver<'a> {
                             kind,
                             width,
                         } => {
-                            let dst_ty = self.types.fetch_if(|t| {
-                                t.inner == crate::TypeInner::Vector { size, kind, width }
-                            });
+                            let dst_ty = self.types.insert(
+                                Type {
+                                    name: None,
+                                    inner: crate::TypeInner::Vector { size, kind, width },
+                                },
+                                span,
+                            );
                             (dst_ty, &src_components[..])
                         }
-                        _ => (None, &[][..]),
+                        _ => {
+                            return Err(ConstantSolvingError::SwizzleVectorOnly);
+                        }
                     },
                 };
-                //TODO: register the new type if needed
-                let ty = ty.ok_or(ConstantSolvingError::DestinationTypeNotFound)?;
+
                 let components = pattern
                     .iter()
                     .map(|&sc| src_components[sc as usize])
@@ -338,7 +352,7 @@ impl<'a> ConstantSolver<'a> {
                 }
 
                 for component in components {
-                    *component = self.cast(*component, kind, target_width, span.clone())?;
+                    *component = self.cast(*component, kind, target_width, span)?;
                 }
             }
         }
@@ -385,7 +399,7 @@ impl<'a> ConstantSolver<'a> {
                 }
 
                 for component in components {
-                    *component = self.unary_op(op, *component, span.clone())?
+                    *component = self.unary_op(op, *component, span)?
                 }
             }
         }
@@ -486,14 +500,14 @@ impl<'a> ConstantSolver<'a> {
             (&ConstantInner::Composite { ref components, ty }, &ConstantInner::Scalar { .. }) => {
                 let mut components = components.clone();
                 for comp in components.iter_mut() {
-                    *comp = self.binary_op(op, *comp, right, span.clone())?;
+                    *comp = self.binary_op(op, *comp, right, span)?;
                 }
                 ConstantInner::Composite { ty, components }
             }
             (&ConstantInner::Scalar { .. }, &ConstantInner::Composite { ref components, ty }) => {
                 let mut components = components.clone();
                 for comp in components.iter_mut() {
-                    *comp = self.binary_op(op, left, *comp, span.clone())?;
+                    *comp = self.binary_op(op, left, *comp, span)?;
                 }
                 ConstantInner::Composite { ty, components }
             }
@@ -517,18 +531,18 @@ mod tests {
 
     use crate::{
         Arena, Constant, ConstantInner, Expression, ScalarKind, ScalarValue, Type, TypeInner,
-        UnaryOperator, VectorSize,
+        UnaryOperator, UniqueArena, VectorSize,
     };
 
     use super::ConstantSolver;
 
     #[test]
     fn unary_op() {
-        let mut types = Arena::new();
+        let mut types = UniqueArena::new();
         let mut expressions = Arena::new();
         let mut constants = Arena::new();
 
-        let vec_ty = types.append(
+        let vec_ty = types.insert(
             Type {
                 name: None,
                 inner: TypeInner::Vector {
@@ -604,7 +618,7 @@ mod tests {
         );
 
         let mut solver = ConstantSolver {
-            types: &types,
+            types: &mut types,
             expressions: &expressions,
             constants: &mut constants,
         };
@@ -684,7 +698,7 @@ mod tests {
         );
 
         let mut solver = ConstantSolver {
-            types: &Arena::new(),
+            types: &mut UniqueArena::new(),
             expressions: &expressions,
             constants: &mut constants,
         };
@@ -702,11 +716,11 @@ mod tests {
 
     #[test]
     fn access() {
-        let mut types = Arena::new();
+        let mut types = UniqueArena::new();
         let mut expressions = Arena::new();
         let mut constants = Arena::new();
 
-        let matrix_ty = types.append(
+        let matrix_ty = types.insert(
             Type {
                 name: None,
                 inner: TypeInner::Matrix {
@@ -718,7 +732,7 @@ mod tests {
             Default::default(),
         );
 
-        let vec_ty = types.append(
+        let vec_ty = types.insert(
             Type {
                 name: None,
                 inner: TypeInner::Vector {
@@ -815,7 +829,7 @@ mod tests {
         );
 
         let mut solver = ConstantSolver {
-            types: &types,
+            types: &mut types,
             expressions: &expressions,
             constants: &mut constants,
         };
