@@ -5,16 +5,41 @@ use super::{
     },
     error::{Error, ErrorKind},
     types::{scalar_components, type_power},
-    Parser, Result, SourceMetadata,
+    Parser, Result,
 };
 use crate::{
     front::{Emitter, Typifier},
     Arena, BinaryOperator, Block, Constant, Expression, FastHashMap, FunctionArgument, Handle,
-    LocalVariable, RelationalFunction, ScalarKind, ScalarValue, Statement, StorageClass, Type,
-    TypeInner, VectorSize,
+    LocalVariable, RelationalFunction, ScalarKind, ScalarValue, Span, Statement, StorageClass,
+    Type, TypeInner, VectorSize,
 };
 use std::{convert::TryFrom, ops::Index};
 
+/// The position at which an expression is, used while lowering
+#[derive(Clone, Copy, PartialEq, Eq)]
+pub enum ExprPos {
+    /// The expression is in the left hand side of an assignment
+    Lhs,
+    /// The expression is in the right hand side of an assignment
+    Rhs,
+    /// The expression is an array being indexed, needed to allow constant
+    /// arrays to be dinamically indexed
+    ArrayBase {
+        /// The index is a constant
+        constant_index: bool,
+    },
+}
+
+impl ExprPos {
+    /// Returns an lhs position if the current position is lhs otherwise ArrayBase
+    fn maybe_array_base(&self, constant_index: bool) -> Self {
+        match *self {
+            ExprPos::Lhs => *self,
+            _ => ExprPos::ArrayBase { constant_index },
+        }
+    }
+}
+
 #[derive(Debug)]
 pub struct Context {
     pub expressions: Arena<Expression>,
@@ -77,51 +102,57 @@ impl Context {
         body: &mut Block,
     ) {
         self.emit_flush(body);
-        let (expr, load) = match kind {
+        let (expr, load, constant) = match kind {
             GlobalLookupKind::Variable(v) => {
-                let span = parser.module.global_variables.get_span(v).clone();
+                let span = parser.module.global_variables.get_span(v);
                 let res = (
                     self.expressions.append(Expression::GlobalVariable(v), span),
                     parser.module.global_variables[v].class != StorageClass::Handle,
+                    None,
                 );
                 self.emit_start();
 
                 res
             }
             GlobalLookupKind::BlockSelect(handle, index) => {
-                let span = parser.module.global_variables.get_span(handle).clone();
+                let span = parser.module.global_variables.get_span(handle);
                 let base = self
                     .expressions
-                    .append(Expression::GlobalVariable(handle), span.clone());
+                    .append(Expression::GlobalVariable(handle), span);
                 self.emit_start();
                 let expr = self
                     .expressions
                     .append(Expression::AccessIndex { base, index }, span);
 
-                (expr, {
-                    let ty = parser.module.global_variables[handle].ty;
+                (
+                    expr,
+                    {
+                        let ty = parser.module.global_variables[handle].ty;
 
-                    match parser.module.types[ty].inner {
-                        TypeInner::Struct { ref members, .. } => {
-                            if let TypeInner::Array {
-                                size: crate::ArraySize::Dynamic,
-                                ..
-                            } = parser.module.types[members[index as usize].ty].inner
-                            {
-                                false
-                            } else {
-                                true
+                        match parser.module.types[ty].inner {
+                            TypeInner::Struct { ref members, .. } => {
+                                if let TypeInner::Array {
+                                    size: crate::ArraySize::Dynamic,
+                                    ..
+                                } = parser.module.types[members[index as usize].ty].inner
+                                {
+                                    false
+                                } else {
+                                    true
+                                }
                             }
+                            _ => true,
                         }
-                        _ => true,
-                    }
-                })
+                    },
+                    None,
+                )
             }
-            GlobalLookupKind::Constant(v) => {
-                let span = parser.module.constants.get_span(v).clone();
+            GlobalLookupKind::Constant(v, ty) => {
+                let span = parser.module.constants.get_span(v);
                 let res = (
                     self.expressions.append(Expression::Constant(v), span),
                     false,
+                    Some((v, ty)),
                 );
                 self.emit_start();
                 res
@@ -132,6 +163,7 @@ impl Context {
             expr,
             load,
             mutable,
+            constant,
             entry_arg,
         };
 
@@ -149,14 +181,14 @@ impl Context {
     pub fn add_expression(
         &mut self,
         expr: Expression,
-        meta: SourceMetadata,
+        meta: Span,
         body: &mut Block,
     ) -> Handle<Expression> {
         let needs_pre_emit = expr.needs_pre_emit();
         if needs_pre_emit {
             self.emit_flush(body);
         }
-        let handle = self.expressions.append(expr, meta.as_span());
+        let handle = self.expressions.append(expr, meta);
         if needs_pre_emit {
             self.emit_start();
         }
@@ -194,6 +226,7 @@ impl Context {
                     expr,
                     load: true,
                     mutable,
+                    constant: None,
                     entry_arg: None,
                 },
             );
@@ -205,7 +238,7 @@ impl Context {
         &mut self,
         parser: &mut Parser,
         body: &mut Block,
-        name_meta: Option<(String, SourceMetadata)>,
+        name_meta: Option<(String, Span)>,
         ty: Handle<Type>,
         qualifier: ParameterQualifier,
     ) {
@@ -223,8 +256,8 @@ impl Context {
         };
 
         if qualifier.is_lhs() {
-            let span = parser.module.types.get_span(arg.ty).clone();
-            arg.ty = parser.module.types.fetch_or_append(
+            let span = parser.module.types.get_span(arg.ty);
+            arg.ty = parser.module.types.insert(
                 Type {
                     name: None,
                     inner: TypeInner::Pointer {
@@ -255,7 +288,7 @@ impl Context {
                         ty,
                         init: None,
                     },
-                    meta.as_span(),
+                    meta,
                 );
                 let local_expr = self.add_expression(Expression::LocalVariable(handle), meta, body);
 
@@ -267,7 +300,7 @@ impl Context {
                         pointer: local_expr,
                         value: expr,
                     },
-                    meta.as_span(),
+                    meta,
                 );
 
                 if let Some(current) = self.scopes.last_mut() {
@@ -277,6 +310,7 @@ impl Context {
                             expr: local_expr,
                             load: true,
                             mutable,
+                            constant: None,
                             entry_arg: None,
                         },
                     );
@@ -288,6 +322,7 @@ impl Context {
                         expr,
                         load,
                         mutable,
+                        constant: None,
                         entry_arg: None,
                     },
                 );
@@ -323,10 +358,10 @@ impl Context {
         mut stmt: StmtContext,
         parser: &mut Parser,
         expr: Handle<HirExpr>,
-        lhs: bool,
+        pos: ExprPos,
         body: &mut Block,
-    ) -> Result<(Option<Handle<Expression>>, SourceMetadata)> {
-        let res = self.lower_inner(&stmt, parser, expr, lhs, body);
+    ) -> Result<(Option<Handle<Expression>>, Span)> {
+        let res = self.lower_inner(&stmt, parser, expr, pos, body);
 
         stmt.hir_exprs.clear();
         self.stmt_ctx = Some(stmt);
@@ -344,10 +379,10 @@ impl Context {
         mut stmt: StmtContext,
         parser: &mut Parser,
         expr: Handle<HirExpr>,
-        lhs: bool,
+        pos: ExprPos,
         body: &mut Block,
-    ) -> Result<(Handle<Expression>, SourceMetadata)> {
-        let res = self.lower_expect_inner(&stmt, parser, expr, lhs, body);
+    ) -> Result<(Handle<Expression>, Span)> {
+        let res = self.lower_expect_inner(&stmt, parser, expr, pos, body);
 
         stmt.hir_exprs.clear();
         self.stmt_ctx = Some(stmt);
@@ -365,10 +400,10 @@ impl Context {
         stmt: &StmtContext,
         parser: &mut Parser,
         expr: Handle<HirExpr>,
-        lhs: bool,
+        pos: ExprPos,
         body: &mut Block,
-    ) -> Result<(Handle<Expression>, SourceMetadata)> {
-        let (maybe_expr, meta) = self.lower_inner(stmt, parser, expr, lhs, body)?;
+    ) -> Result<(Handle<Expression>, Span)> {
+        let (maybe_expr, meta) = self.lower_inner(stmt, parser, expr, pos, body)?;
 
         let expr = match maybe_expr {
             Some(e) => e,
@@ -389,20 +424,33 @@ impl Context {
         stmt: &StmtContext,
         parser: &mut Parser,
         expr: Handle<HirExpr>,
-        lhs: bool,
+        pos: ExprPos,
         body: &mut Block,
-    ) -> Result<(Option<Handle<Expression>>, SourceMetadata)> {
+    ) -> Result<(Option<Handle<Expression>>, Span)> {
         let HirExpr { ref kind, meta } = stmt.hir_exprs[expr];
 
         let handle = match *kind {
             HirExprKind::Access { base, index } => {
-                let base = self.lower_expect_inner(stmt, parser, base, true, body)?.0;
                 let (index, index_meta) =
-                    self.lower_expect_inner(stmt, parser, index, false, body)?;
+                    self.lower_expect_inner(stmt, parser, index, ExprPos::Rhs, body)?;
+                let maybe_constant_index = match pos {
+                    // Don't try to generate `AccessIndex` if in a LHS position, since it
+                    // wouldn't produce a pointer.
+                    ExprPos::Lhs => None,
+                    _ => parser.solve_constant(self, index, index_meta).ok(),
+                };
 
-                let pointer = parser
-                    .solve_constant(self, index, index_meta)
-                    .ok()
+                let base = self
+                    .lower_expect_inner(
+                        stmt,
+                        parser,
+                        base,
+                        pos.maybe_array_base(maybe_constant_index.is_some()),
+                        body,
+                    )?
+                    .0;
+
+                let pointer = maybe_constant_index
                     .and_then(|constant| {
                         Some(self.add_expression(
                             Expression::AccessIndex {
@@ -427,7 +475,7 @@ impl Context {
                         self.add_expression(Expression::Access { base, index }, meta, body)
                     });
 
-                if !lhs {
+                if ExprPos::Rhs == pos {
                     let resolved = parser.resolve_type(self, pointer, meta)?;
                     if resolved.pointer_class().is_some() {
                         return Ok((
@@ -440,18 +488,18 @@ impl Context {
                 pointer
             }
             HirExprKind::Select { base, ref field } => {
-                let base = self.lower_expect_inner(stmt, parser, base, lhs, body)?.0;
+                let base = self.lower_expect_inner(stmt, parser, base, pos, body)?.0;
 
-                parser.field_selection(self, lhs, body, base, field, meta)?
+                parser.field_selection(self, ExprPos::Lhs == pos, body, base, field, meta)?
             }
-            HirExprKind::Constant(constant) if !lhs => {
+            HirExprKind::Constant(constant) if pos == ExprPos::Rhs => {
                 self.add_expression(Expression::Constant(constant), meta, body)
             }
-            HirExprKind::Binary { left, op, right } if !lhs => {
+            HirExprKind::Binary { left, op, right } if pos == ExprPos::Rhs => {
                 let (mut left, left_meta) =
-                    self.lower_expect_inner(stmt, parser, left, false, body)?;
+                    self.lower_expect_inner(stmt, parser, left, pos, body)?;
                 let (mut right, right_meta) =
-                    self.lower_expect_inner(stmt, parser, right, false, body)?;
+                    self.lower_expect_inner(stmt, parser, right, pos, body)?;
 
                 match op {
                     BinaryOperator::ShiftLeft | BinaryOperator::ShiftRight => self
@@ -480,7 +528,7 @@ impl Context {
 
                             let argument = self
                                 .expressions
-                                .append(Expression::Binary { op, left, right }, meta.as_span());
+                                .append(Expression::Binary { op, left, right }, meta);
 
                             self.add_expression(
                                 Expression::Relational { fun, argument },
@@ -496,6 +544,9 @@ impl Context {
                         BinaryOperator::Add
                         | BinaryOperator::Subtract
                         | BinaryOperator::Divide
+                        | BinaryOperator::And
+                        | BinaryOperator::ExclusiveOr
+                        | BinaryOperator::InclusiveOr
                         | BinaryOperator::ShiftLeft
                         | BinaryOperator::ShiftRight => {
                             let scalar_vector = self.add_expression(
@@ -519,7 +570,12 @@ impl Context {
                         }
                     },
                     (&TypeInner::Scalar { .. }, &TypeInner::Vector { size, .. }) => match op {
-                        BinaryOperator::Add | BinaryOperator::Subtract | BinaryOperator::Divide => {
+                        BinaryOperator::Add
+                        | BinaryOperator::Subtract
+                        | BinaryOperator::Divide
+                        | BinaryOperator::And
+                        | BinaryOperator::ExclusiveOr
+                        | BinaryOperator::InclusiveOr => {
                             let scalar_vector = self.add_expression(
                                 Expression::Splat { size, value: left },
                                 meta,
@@ -543,13 +599,13 @@ impl Context {
                     _ => self.add_expression(Expression::Binary { left, op, right }, meta, body),
                 }
             }
-            HirExprKind::Unary { op, expr } if !lhs => {
-                let expr = self.lower_expect_inner(stmt, parser, expr, false, body)?.0;
+            HirExprKind::Unary { op, expr } if pos == ExprPos::Rhs => {
+                let expr = self.lower_expect_inner(stmt, parser, expr, pos, body)?.0;
 
                 self.add_expression(Expression::Unary { op, expr }, meta, body)
             }
-            HirExprKind::Variable(ref var) => {
-                if lhs {
+            HirExprKind::Variable(ref var) => match pos {
+                ExprPos::Lhs => {
                     if !var.mutable {
                         parser.errors.push(Error {
                             kind: ErrorKind::SemanticError(
@@ -559,14 +615,32 @@ impl Context {
                         })
                     }
 
-                    var.expr
-                } else if var.load {
-                    self.add_expression(Expression::Load { pointer: var.expr }, meta, body)
-                } else {
                     var.expr
                 }
-            }
-            HirExprKind::Call(ref call) if !lhs => {
+                ExprPos::ArrayBase {
+                    constant_index: false,
+                } => {
+                    if let Some((constant, ty)) = var.constant {
+                        let local = self.locals.append(
+                            LocalVariable {
+                                name: None,
+                                ty,
+                                init: Some(constant),
+                            },
+                            Span::default(),
+                        );
+
+                        self.add_expression(Expression::LocalVariable(local), Span::default(), body)
+                    } else {
+                        var.expr
+                    }
+                }
+                _ if var.load => {
+                    self.add_expression(Expression::Load { pointer: var.expr }, meta, body)
+                }
+                _ => var.expr,
+            },
+            HirExprKind::Call(ref call) if pos != ExprPos::Lhs => {
                 let maybe_expr = parser.function_or_constructor_call(
                     self,
                     stmt,
@@ -581,14 +655,14 @@ impl Context {
                 condition,
                 accept,
                 reject,
-            } if !lhs => {
+            } if ExprPos::Lhs != pos => {
                 let condition = self
-                    .lower_expect_inner(stmt, parser, condition, false, body)?
+                    .lower_expect_inner(stmt, parser, condition, ExprPos::Rhs, body)?
                     .0;
                 let (mut accept, accept_meta) =
-                    self.lower_expect_inner(stmt, parser, accept, false, body)?;
+                    self.lower_expect_inner(stmt, parser, accept, pos, body)?;
                 let (mut reject, reject_meta) =
-                    self.lower_expect_inner(stmt, parser, reject, false, body)?;
+                    self.lower_expect_inner(stmt, parser, reject, pos, body)?;
 
                 self.binary_implicit_conversion(
                     parser,
@@ -608,10 +682,11 @@ impl Context {
                     body,
                 )
             }
-            HirExprKind::Assign { tgt, value } if !lhs => {
-                let (pointer, ptr_meta) = self.lower_expect_inner(stmt, parser, tgt, true, body)?;
+            HirExprKind::Assign { tgt, value } if ExprPos::Lhs != pos => {
+                let (pointer, ptr_meta) =
+                    self.lower_expect_inner(stmt, parser, tgt, ExprPos::Lhs, body)?;
                 let (mut value, value_meta) =
-                    self.lower_expect_inner(stmt, parser, value, false, body)?;
+                    self.lower_expect_inner(stmt, parser, value, ExprPos::Rhs, body)?;
 
                 let scalar_components = self.expr_scalar_components(parser, pointer, ptr_meta)?;
 
@@ -664,112 +739,120 @@ impl Context {
                                 pointer: dst,
                                 value: src,
                             },
-                            meta.as_span(),
+                            meta,
                         );
                     }
                 } else {
                     self.emit_flush(body);
                     self.emit_start();
 
-                    body.push(Statement::Store { pointer, value }, meta.as_span());
+                    body.push(Statement::Store { pointer, value }, meta);
                 }
 
                 value
             }
-            HirExprKind::IncDec {
-                increment,
-                postfix,
-                expr,
-            } => {
-                let op = match increment {
-                    true => BinaryOperator::Add,
-                    false => BinaryOperator::Subtract,
-                };
-
-                let pointer = self.lower_expect_inner(stmt, parser, expr, true, body)?.0;
+            HirExprKind::PrePostfix { op, postfix, expr } if ExprPos::Lhs != pos => {
+                let pointer = self
+                    .lower_expect_inner(stmt, parser, expr, ExprPos::Lhs, body)?
+                    .0;
                 let left = self.add_expression(Expression::Load { pointer }, meta, body);
 
-                let uint = match parser.resolve_type(self, left, meta)?.scalar_kind() {
-                    Some(ScalarKind::Sint) => false,
-                    Some(ScalarKind::Uint) => true,
-                    _ => {
+                let make_constant_inner = |kind, width| {
+                    let value = match kind {
+                        ScalarKind::Sint => crate::ScalarValue::Sint(1),
+                        ScalarKind::Uint => crate::ScalarValue::Uint(1),
+                        ScalarKind::Float => crate::ScalarValue::Float(1.0),
+                        ScalarKind::Bool => return None,
+                    };
+
+                    Some(crate::ConstantInner::Scalar { width, value })
+                };
+                let res = match *parser.resolve_type(self, left, meta)? {
+                    TypeInner::Scalar { kind, width } => {
+                        let ty = TypeInner::Scalar { kind, width };
+                        make_constant_inner(kind, width).map(|i| (ty, i, None, None))
+                    }
+                    TypeInner::Vector { size, kind, width } => {
+                        let ty = TypeInner::Vector { size, kind, width };
+                        make_constant_inner(kind, width).map(|i| (ty, i, Some(size), None))
+                    }
+                    TypeInner::Matrix {
+                        columns,
+                        rows,
+                        width,
+                    } => {
+                        let ty = TypeInner::Matrix {
+                            columns,
+                            rows,
+                            width,
+                        };
+                        make_constant_inner(ScalarKind::Float, width)
+                            .map(|i| (ty, i, Some(rows), Some(columns)))
+                    }
+                    _ => None,
+                };
+                let (ty_inner, inner, rows, columns) = match res {
+                    Some(res) => res,
+                    None => {
                         parser.errors.push(Error {
                             kind: ErrorKind::SemanticError(
-                                "Increment/decrement operations must operate in integers".into(),
+                                "Increment/decrement only works on scalar/vector/matrix".into(),
                             ),
                             meta,
                         });
-                        true
+                        return Ok((Some(left), meta));
                     }
                 };
 
-                let one = parser.module.constants.append(
+                let constant_1 = parser.module.constants.append(
                     Constant {
                         name: None,
                         specialization: None,
-                        inner: crate::ConstantInner::Scalar {
-                            width: 4,
-                            value: match uint {
-                                true => crate::ScalarValue::Uint(1),
-                                false => crate::ScalarValue::Sint(1),
-                            },
-                        },
+                        inner,
                     },
                     Default::default(),
                 );
-                let right = self.add_expression(Expression::Constant(one), meta, body);
+                let mut right = self.add_expression(Expression::Constant(constant_1), meta, body);
+
+                // Glsl allows pre/postfixes operations on vectors and matrices, so if the
+                // target is either of them change the right side of the addition to be splatted
+                // to the same size as the target, furthermore if the target is a matrix
+                // use a composed matrix using the splatted value.
+                if let Some(size) = rows {
+                    right =
+                        self.add_expression(Expression::Splat { size, value: right }, meta, body);
+
+                    if let Some(cols) = columns {
+                        let ty = parser.module.types.insert(
+                            Type {
+                                name: None,
+                                inner: ty_inner,
+                            },
+                            meta,
+                        );
+
+                        right = self.add_expression(
+                            Expression::Compose {
+                                ty,
+                                components: std::iter::repeat(right).take(cols as usize).collect(),
+                            },
+                            meta,
+                            body,
+                        );
+                    }
+                }
 
                 let value = self.add_expression(Expression::Binary { op, left, right }, meta, body);
 
+                self.emit_flush(body);
+                self.emit_start();
+
+                body.push(Statement::Store { pointer, value }, meta);
+
                 if postfix {
-                    let local = self.locals.append(
-                        LocalVariable {
-                            name: None,
-                            ty: parser.module.types.fetch_or_append(
-                                Type {
-                                    name: None,
-                                    inner: TypeInner::Scalar {
-                                        kind: match uint {
-                                            true => ScalarKind::Uint,
-                                            false => ScalarKind::Sint,
-                                        },
-                                        width: 4,
-                                    },
-                                },
-                                meta.as_span(),
-                            ),
-                            init: None,
-                        },
-                        meta.as_span(),
-                    );
-
-                    let expr = self.add_expression(Expression::LocalVariable(local), meta, body);
-                    let load = self.add_expression(Expression::Load { pointer: expr }, meta, body);
-
-                    self.emit_flush(body);
-                    self.emit_start();
-
-                    body.push(
-                        Statement::Store {
-                            pointer: expr,
-                            value: left,
-                        },
-                        meta.as_span(),
-                    );
-
-                    self.emit_flush(body);
-                    self.emit_start();
-
-                    body.push(Statement::Store { pointer, value }, meta.as_span());
-
-                    load
-                } else {
-                    self.emit_flush(body);
-                    self.emit_start();
-
-                    body.push(Statement::Store { pointer, value }, meta.as_span());
-
                     left
+                } else {
+                    value
                 }
             }
             _ => {
@@ -790,7 +873,7 @@ impl Context {
         &mut self,
         parser: &mut Parser,
         expr: Handle<Expression>,
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Result<Option<(ScalarKind, crate::Bytes)>> {
         let ty = parser.resolve_type(self, expr, meta)?;
         Ok(scalar_components(ty))
@@ -800,18 +883,37 @@ impl Context {
         &mut self,
         parser: &mut Parser,
         expr: Handle<Expression>,
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Result<Option<u32>> {
         Ok(self
             .expr_scalar_components(parser, expr, meta)?
             .and_then(|(kind, width)| type_power(kind, width)))
     }
 
+    pub fn conversion(
+        &mut self,
+        expr: &mut Handle<Expression>,
+        meta: Span,
+        kind: ScalarKind,
+        width: crate::Bytes,
+    ) -> Result<()> {
+        *expr = self.expressions.append(
+            Expression::As {
+                expr: *expr,
+                kind,
+                convert: Some(width),
+            },
+            meta,
+        );
+
+        Ok(())
+    }
+
     pub fn implicit_conversion(
         &mut self,
         parser: &mut Parser,
         expr: &mut Handle<Expression>,
-        meta: SourceMetadata,
+        meta: Span,
         kind: ScalarKind,
         width: crate::Bytes,
     ) -> Result<()> {
@@ -820,14 +922,7 @@ impl Context {
             self.expr_power(parser, *expr, meta)?,
         ) {
             if tgt_power > expr_power {
-                *expr = self.expressions.append(
-                    Expression::As {
-                        expr: *expr,
-                        kind,
-                        convert: Some(width),
-                    },
-                    meta.as_span(),
-                )
+                self.conversion(expr, meta, kind, width)?;
             }
         }
 
@@ -838,9 +933,9 @@ impl Context {
         &mut self,
         parser: &mut Parser,
         left: &mut Handle<Expression>,
-        left_meta: SourceMetadata,
+        left_meta: Span,
         right: &mut Handle<Expression>,
-        right_meta: SourceMetadata,
+        right_meta: Span,
     ) -> Result<()> {
         let left_components = self.expr_scalar_components(parser, *left, left_meta)?;
         let right_components = self.expr_scalar_components(parser, *right, right_meta)?;
@@ -855,25 +950,11 @@ impl Context {
         ) {
             match left_power.cmp(&right_power) {
                 std::cmp::Ordering::Less => {
-                    *left = self.expressions.append(
-                        Expression::As {
-                            expr: *left,
-                            kind: right_kind,
-                            convert: Some(right_width),
-                        },
-                        left_meta.as_span(),
-                    )
+                    self.conversion(left, left_meta, right_kind, right_width)?;
                 }
                 std::cmp::Ordering::Equal => {}
                 std::cmp::Ordering::Greater => {
-                    *right = self.expressions.append(
-                        Expression::As {
-                            expr: *right,
-                            kind: left_kind,
-                            convert: Some(left_width),
-                        },
-                        right_meta.as_span(),
-                    )
+                    self.conversion(right, right_meta, left_kind, left_width)?;
                 }
             }
         }
@@ -885,7 +966,7 @@ impl Context {
         &mut self,
         parser: &mut Parser,
         expr: &mut Handle<Expression>,
-        meta: SourceMetadata,
+        meta: Span,
         vector_size: Option<VectorSize>,
     ) -> Result<()> {
         let expr_type = parser.resolve_type(self, *expr, meta)?;
@@ -893,7 +974,7 @@ impl Context {
         if let (&TypeInner::Scalar { .. }, Some(size)) = (expr_type, vector_size) {
             *expr = self
                 .expressions
-                .append(Expression::Splat { size, value: *expr }, meta.as_span())
+                .append(Expression::Splat { size, value: *expr }, meta)
         }
 
         Ok(())
@@ -903,7 +984,7 @@ impl Context {
         &mut self,
         size: VectorSize,
         vector: Handle<Expression>,
-        meta: SourceMetadata,
+        meta: Span,
         body: &mut Block,
     ) -> Handle<Expression> {
         self.add_expression(
