@@ -2,12 +2,12 @@ use super::{
     ast::*,
     context::Context,
     error::{Error, ErrorKind},
-    Parser, Result, SourceMetadata,
+    Parser, Result, Span,
 };
 use crate::{
     Binding, Block, BuiltIn, Constant, Expression, GlobalVariable, Handle, Interpolation,
-    LocalVariable, ResourceBinding, ScalarKind, StorageAccess, StorageClass, SwizzleComponent,
-    Type, TypeInner, VectorSize,
+    LocalVariable, ResourceBinding, ScalarKind, ShaderStage, StorageAccess, StorageClass,
+    SwizzleComponent, Type, TypeInner, VectorSize,
 };
 
 macro_rules! qualifier_arm {
@@ -24,11 +24,23 @@ macro_rules! qualifier_arm {
 }
 
 pub struct VarDeclaration<'a> {
-    pub qualifiers: &'a [(TypeQualifier, SourceMetadata)],
+    pub qualifiers: &'a [(TypeQualifier, Span)],
     pub ty: Handle<Type>,
     pub name: Option<String>,
     pub init: Option<Handle<Constant>>,
-    pub meta: SourceMetadata,
+    pub meta: Span,
+}
+
+/// Information about a builtin used in [`add_builtin`](Parser::add_builtin)
+struct BuiltInData {
+    /// The type of the builtin
+    inner: TypeInner,
+    /// The builtin class associated with
+    builtin: BuiltIn,
+    /// Wether it should be allowed to write to the builtin or not
+    mutable: bool,
+    /// The storage used for the builtin
+    storage: StorageQualifier,
 }
 
 pub enum GlobalOrConstant {
@@ -37,12 +49,72 @@ pub enum GlobalOrConstant {
 }
 
 impl Parser {
+    /// Adds a builtin and returns a variable reference to it
+    fn add_builtin(
+        &mut self,
+        ctx: &mut Context,
+        body: &mut Block,
+        name: &str,
+        data: BuiltInData,
+        meta: Span,
+    ) -> Option<VariableReference> {
+        let ty = self.module.types.insert(
+            Type {
+                name: None,
+                inner: data.inner,
+            },
+            meta,
+        );
+
+        let handle = self.module.global_variables.append(
+            GlobalVariable {
+                name: Some(name.into()),
+                class: StorageClass::Private,
+                binding: None,
+                ty,
+                init: None,
+            },
+            meta,
+        );
+
+        let idx = self.entry_args.len();
+        self.entry_args.push(EntryArg {
+            name: None,
+            binding: Binding::BuiltIn(data.builtin),
+            handle,
+            storage: data.storage,
+        });
+
+        self.global_variables.push((
+            name.into(),
+            GlobalLookup {
+                kind: GlobalLookupKind::Variable(handle),
+                entry_arg: Some(idx),
+                mutable: data.mutable,
+            },
+        ));
+
+        let expr = ctx.add_expression(Expression::GlobalVariable(handle), meta, body);
+        ctx.lookup_global_var_exps.insert(
+            name.into(),
+            VariableReference {
+                expr,
+                load: true,
+                mutable: data.mutable,
+                constant: None,
+                entry_arg: Some(idx),
+            },
+        );
+
+        ctx.lookup_global_var(name)
+    }
+
     pub(crate) fn lookup_variable(
         &mut self,
         ctx: &mut Context,
         body: &mut Block,
         name: &str,
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Option<VariableReference> {
         if let Some(local_var) = ctx.lookup_local_var(name) {
             return Some(local_var);
@@ -51,141 +123,122 @@ impl Parser {
             return Some(global_var);
         }
 
-        let mut add_builtin = |inner, builtin, mutable, storage| {
-            let ty = self
-                .module
-                .types
-                .fetch_or_append(Type { name: None, inner }, meta.as_span());
-
-            let handle = self.module.global_variables.append(
-                GlobalVariable {
-                    name: Some(name.into()),
-                    class: StorageClass::Private,
-                    binding: None,
-                    ty,
-                    init: None,
-                },
-                meta.as_span(),
-            );
-
-            let idx = self.entry_args.len();
-            self.entry_args.push(EntryArg {
-                name: None,
-                binding: Binding::BuiltIn(builtin),
-                handle,
-                storage,
-            });
-
-            self.global_variables.push((
-                name.into(),
-                GlobalLookup {
-                    kind: GlobalLookupKind::Variable(handle),
-                    entry_arg: Some(idx),
-                    mutable,
-                },
-            ));
-
-            let expr = ctx.add_expression(Expression::GlobalVariable(handle), meta, body);
-            ctx.lookup_global_var_exps.insert(
-                name.into(),
-                VariableReference {
-                    expr,
-                    load: true,
-                    mutable,
-                    entry_arg: Some(idx),
-                },
-            );
-
-            ctx.lookup_global_var(name)
-        };
-        match name {
-            "gl_Position" => add_builtin(
-                TypeInner::Vector {
+        let data = match name {
+            "gl_Position" => BuiltInData {
+                inner: TypeInner::Vector {
                     size: VectorSize::Quad,
                     kind: ScalarKind::Float,
                     width: 4,
                 },
-                BuiltIn::Position,
-                true,
-                StorageQualifier::Output,
-            ),
-            "gl_FragCoord" => add_builtin(
-                TypeInner::Vector {
+                builtin: BuiltIn::Position,
+                mutable: true,
+                storage: StorageQualifier::Output,
+            },
+            "gl_FragCoord" => BuiltInData {
+                inner: TypeInner::Vector {
                     size: VectorSize::Quad,
                     kind: ScalarKind::Float,
                     width: 4,
                 },
-                BuiltIn::Position,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_FragDepth" => add_builtin(
-                TypeInner::Scalar {
-                    kind: ScalarKind::Float,
-                    width: 4,
-                },
-                BuiltIn::FragDepth,
-                true,
-                StorageQualifier::Output,
-            ),
-            "gl_VertexIndex" => add_builtin(
-                TypeInner::Scalar {
-                    kind: ScalarKind::Uint,
-                    width: 4,
-                },
-                BuiltIn::VertexIndex,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_InstanceIndex" => add_builtin(
-                TypeInner::Scalar {
-                    kind: ScalarKind::Uint,
-                    width: 4,
-                },
-                BuiltIn::InstanceIndex,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_GlobalInvocationID" => add_builtin(
-                TypeInner::Vector {
+                builtin: BuiltIn::Position,
+                mutable: false,
+                storage: StorageQualifier::Input,
+            },
+            "gl_GlobalInvocationID"
+            | "gl_NumWorkGroups"
+            | "gl_WorkGroupSize"
+            | "gl_WorkGroupID"
+            | "gl_LocalInvocationID" => BuiltInData {
+                inner: TypeInner::Vector {
                     size: VectorSize::Tri,
                     kind: ScalarKind::Uint,
                     width: 4,
                 },
-                BuiltIn::GlobalInvocationId,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_NumWorkGroups" => add_builtin(
-                TypeInner::Vector {
-                    size: VectorSize::Tri,
-                    kind: ScalarKind::Uint,
-                    width: 4,
+                builtin: match name {
+                    "gl_GlobalInvocationID" => BuiltIn::GlobalInvocationId,
+                    "gl_NumWorkGroups" => BuiltIn::NumWorkGroups,
+                    "gl_WorkGroupSize" => BuiltIn::WorkGroupSize,
+                    "gl_WorkGroupID" => BuiltIn::WorkGroupId,
+                    "gl_LocalInvocationID" => BuiltIn::LocalInvocationId,
+                    _ => unreachable!(),
                 },
-                BuiltIn::NumWorkGroups,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_FrontFacing" => add_builtin(
-                TypeInner::Scalar {
+                mutable: false,
+                storage: StorageQualifier::Input,
+            },
+            "gl_FrontFacing" => BuiltInData {
+                inner: TypeInner::Scalar {
                     kind: ScalarKind::Bool,
                     width: crate::BOOL_WIDTH,
                 },
-                BuiltIn::FrontFacing,
-                false,
-                StorageQualifier::Input,
-            ),
-            "gl_PrimitiveID" => add_builtin(
-                TypeInner::Scalar {
-                    kind: ScalarKind::Uint,
+                builtin: BuiltIn::FrontFacing,
+                mutable: false,
+                storage: StorageQualifier::Input,
+            },
+            "gl_PointSize" | "gl_FragDepth" => BuiltInData {
+                inner: TypeInner::Scalar {
+                    kind: ScalarKind::Float,
                     width: 4,
                 },
-                BuiltIn::PrimitiveIndex,
-                false,
-                StorageQualifier::Input,
-            ),
-            _ => None,
-        }
+                builtin: match name {
+                    "gl_PointSize" => BuiltIn::PointSize,
+                    "gl_FragDepth" => BuiltIn::FragDepth,
+                    _ => unreachable!(),
+                },
+                mutable: true,
+                storage: StorageQualifier::Output,
+            },
+            "gl_ClipDistance" | "gl_CullDistance" => {
+                let base = self.module.types.insert(
+                    Type {
+                        name: None,
+                        inner: TypeInner::Scalar {
+                            kind: ScalarKind::Float,
+                            width: 4,
+                        },
+                    },
+                    meta,
+                );
+
+                BuiltInData {
+                    inner: TypeInner::Array {
+                        base,
+                        size: crate::ArraySize::Dynamic,
+                        stride: 4,
+                    },
+                    builtin: match name {
+                        "gl_ClipDistance" => BuiltIn::PointSize,
+                        "gl_CullDistance" => BuiltIn::FragDepth,
+                        _ => unreachable!(),
+                    },
+                    mutable: self.meta.stage == ShaderStage::Vertex,
+                    storage: StorageQualifier::Output,
+                }
+            }
+            _ => {
+                let builtin = match name {
+                    "gl_BaseVertex" => BuiltIn::BaseVertex,
+                    "gl_BaseInstance" => BuiltIn::BaseInstance,
+                    "gl_PrimitiveID" => BuiltIn::PrimitiveIndex,
+                    "gl_InstanceIndex" => BuiltIn::InstanceIndex,
+                    "gl_VertexIndex" => BuiltIn::VertexIndex,
+                    "gl_SampleID" => BuiltIn::SampleIndex,
+                    "gl_LocalInvocationIndex" => BuiltIn::LocalInvocationIndex,
+                    _ => return None,
+                };
+
+                BuiltInData {
+                    inner: TypeInner::Scalar {
+                        kind: ScalarKind::Uint,
+                        width: 4,
+                    },
+                    builtin,
+                    mutable: false,
+                    storage: StorageQualifier::Input,
+                }
+            }
+        };
+
+        self.add_builtin(ctx, body, name, data, meta)
     }
 
     pub(crate) fn field_selection(
@@ -195,7 +248,7 @@ impl Parser {
         body: &mut Block,
         expression: Handle<Expression>,
         name: &str,
-        meta: SourceMetadata,
+        meta: Span,
     ) -> Result<Handle<Expression>> {
         let (ty, is_pointer) = match *self.resolve_type(ctx, expression, meta)? {
             TypeInner::Pointer { base, .. } => (&self.module.types[base].inner, true),
@@ -509,7 +562,7 @@ impl Parser {
                     ty,
                     init,
                 },
-                meta.as_span(),
+                meta,
             );
 
             let idx = self.entry_args.len();
@@ -543,7 +596,7 @@ impl Parser {
             })?;
             if let Some(name) = name {
                 let lookup = GlobalLookup {
-                    kind: GlobalLookupKind::Constant(init),
+                    kind: GlobalLookupKind::Constant(init, ty),
                     entry_arg: None,
                     mutable: false,
                 };
@@ -580,7 +633,7 @@ impl Parser {
                 ty,
                 init,
             },
-            meta.as_span(),
+            meta,
         );
 
         if let Some(name) = name {
@@ -658,7 +711,7 @@ impl Parser {
                 ty,
                 init,
             },
-            meta.as_span(),
+            meta,
         );
         let expr = ctx.add_expression(Expression::LocalVariable(handle), meta, body);
 
