@@ -1,5 +1,5 @@
 use crate::{
-    arena::{Arena, Handle},
+    arena::{Arena, Handle, UniqueArena},
     FunctionArgument,
 };
 
@@ -62,7 +62,7 @@ fn extract_image_coordinates(
     extra_coordinate: ExtraCoordinate,
     base: Handle<crate::Expression>,
     coordinate_ty: Handle<crate::Type>,
-    type_arena: &Arena<crate::Type>,
+    type_arena: &UniqueArena<crate::Type>,
     expressions: &mut Arena<crate::Expression>,
 ) -> (Handle<crate::Expression>, Option<Handle<crate::Expression>>) {
     let (given_size, kind) = match type_arena[coordinate_ty].inner {
@@ -74,13 +74,13 @@ fn extract_image_coordinates(
     let required_size = image_dim.required_coordinate_size();
     let required_ty = required_size.map(|size| {
         type_arena
-            .fetch_if(|ty| {
-                ty.inner
-                    == crate::TypeInner::Vector {
-                        size,
-                        kind,
-                        width: 4,
-                    }
+            .get(&crate::Type {
+                name: None,
+                inner: crate::TypeInner::Vector {
+                    size,
+                    kind,
+                    width: 4,
+                },
             })
             .expect("Required coordinate type should have been set up by `parse_type_image`!")
     });
@@ -89,22 +89,19 @@ fn extract_image_coordinates(
         index: required_size.map_or(1, |size| size as u32),
     };
 
-    let base_span = expressions.get_span(base).clone();
+    let base_span = expressions.get_span(base);
 
     match extra_coordinate {
         ExtraCoordinate::ArrayLayer => {
             let extracted = match required_size {
-                None => expressions.append(
-                    crate::Expression::AccessIndex { base, index: 0 },
-                    base_span.clone(),
-                ),
+                None => {
+                    expressions.append(crate::Expression::AccessIndex { base, index: 0 }, base_span)
+                }
                 Some(size) => {
                     let mut components = Vec::with_capacity(size as usize);
                     for index in 0..size as u32 {
-                        let comp = expressions.append(
-                            crate::Expression::AccessIndex { base, index },
-                            base_span.clone(),
-                        );
+                        let comp = expressions
+                            .append(crate::Expression::AccessIndex { base, index }, base_span);
                         components.push(comp);
                     }
                     expressions.append(
@@ -112,11 +109,11 @@ fn extract_image_coordinates(
                             ty: required_ty.unwrap(),
                             components,
                         },
-                        base_span.clone(),
+                        base_span,
                     )
                 }
             };
-            let array_index_f32 = expressions.append(extra_expr, base_span.clone());
+            let array_index_f32 = expressions.append(extra_expr, base_span);
             let array_index = expressions.append(
                 crate::Expression::As {
                     kind: crate::ScalarKind::Sint,
@@ -128,13 +125,11 @@ fn extract_image_coordinates(
             (extracted, Some(array_index))
         }
         ExtraCoordinate::Projection => {
-            let projection = expressions.append(extra_expr, base_span.clone());
+            let projection = expressions.append(extra_expr, base_span);
             let divided = match required_size {
                 None => {
-                    let temp = expressions.append(
-                        crate::Expression::AccessIndex { base, index: 0 },
-                        base_span.clone(),
-                    );
+                    let temp = expressions
+                        .append(crate::Expression::AccessIndex { base, index: 0 }, base_span);
                     expressions.append(
                         crate::Expression::Binary {
                             op: crate::BinaryOperator::Divide,
@@ -147,17 +142,15 @@ fn extract_image_coordinates(
                 Some(size) => {
                     let mut components = Vec::with_capacity(size as usize);
                     for index in 0..size as u32 {
-                        let temp = expressions.append(
-                            crate::Expression::AccessIndex { base, index },
-                            base_span.clone(),
-                        );
+                        let temp = expressions
+                            .append(crate::Expression::AccessIndex { base, index }, base_span);
                         let comp = expressions.append(
                             crate::Expression::Binary {
                                 op: crate::BinaryOperator::Divide,
                                 left: temp,
                                 right: projection,
                             },
-                            base_span.clone(),
+                            base_span,
                         );
                         components.push(comp);
                     }
@@ -191,7 +184,7 @@ fn extract_image_coordinates(
 pub(super) fn patch_comparison_type(
     flags: SamplingFlags,
     var: &mut crate::GlobalVariable,
-    arena: &mut Arena<crate::Type>,
+    arena: &mut UniqueArena<crate::Type>,
 ) -> bool {
     if !flags.contains(SamplingFlags::COMPARISON) {
         return true;
@@ -202,7 +195,7 @@ pub(super) fn patch_comparison_type(
 
     log::debug!("Flipping comparison for {:?}", var);
     let original_ty = &arena[var.ty];
-    let original_ty_span = arena.get_span(var.ty).clone();
+    let original_ty_span = arena.get_span(var.ty);
     let ty_inner = match original_ty.inner {
         crate::TypeInner::Image {
             class: crate::ImageClass::Sampled { multi, .. },
@@ -218,7 +211,7 @@ pub(super) fn patch_comparison_type(
     };
 
     let name = original_ty.name.clone();
-    var.ty = arena.append(
+    var.ty = arena.insert(
         crate::Type {
             name,
             inner: ty_inner,
@@ -246,7 +239,7 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         Ok(())
     }
 
-    pub(super) fn parse_image_uncouple(&mut self) -> Result<(), Error> {
+    pub(super) fn parse_image_uncouple(&mut self, block_id: spirv::Word) -> Result<(), Error> {
         let result_type_id = self.next()?;
         let result_id = self.next()?;
         let sampled_image_id = self.next()?;
@@ -255,6 +248,7 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             LookupExpression {
                 handle: self.lookup_sampled_image.lookup(sampled_image_id)?.image,
                 type_id: result_type_id,
+                block_id,
             },
         );
         Ok(())
@@ -263,7 +257,7 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
     pub(super) fn parse_image_write(
         &mut self,
         words_left: u16,
-        type_arena: &Arena<crate::Type>,
+        type_arena: &UniqueArena<crate::Type>,
         global_arena: &Arena<crate::GlobalVariable>,
         arguments: &[FunctionArgument],
         expressions: &mut Arena<crate::Expression>,
@@ -320,10 +314,11 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
     pub(super) fn parse_image_load(
         &mut self,
         mut words_left: u16,
-        type_arena: &Arena<crate::Type>,
-        global_arena: &Arena<crate::GlobalVariable>,
-        arguments: &[FunctionArgument],
-        expressions: &mut Arena<crate::Expression>,
+        ctx: &mut super::BlockContext,
+        emitter: &mut crate::front::Emitter,
+        block: &mut crate::Block,
+        block_id: spirv::Word,
+        body_idx: usize,
     ) -> Result<(), Error> {
         let start = self.data_offset;
         let result_type_id = self.next()?;
@@ -365,12 +360,20 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             image_ops ^= bit;
         }
 
+        // No need to call get_expr_handle here since only globals/arguments are
+        // allowed as images and they are always in the root scope
         let image_lexp = self.lookup_expression.lookup(image_id)?;
-        let image_ty = expressions.get_image_expr_ty(image_lexp.handle, global_arena, arguments)?;
+        let image_ty = ctx.expressions.get_image_expr_ty(
+            image_lexp.handle,
+            ctx.global_arena,
+            ctx.arguments,
+        )?;
 
         let coord_lexp = self.lookup_expression.lookup(coordinate_id)?;
+        let coord_handle =
+            self.get_expr_handle(coordinate_id, coord_lexp, ctx, emitter, block, body_idx);
         let coord_type_handle = self.lookup_type.lookup(coord_lexp.type_id)?.handle;
-        let (coordinate, array_index) = match type_arena[image_ty].inner {
+        let (coordinate, array_index) = match ctx.type_arena[image_ty].inner {
             crate::TypeInner::Image {
                 dim,
                 arrayed,
@@ -382,10 +385,10 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
                 } else {
                     ExtraCoordinate::Garbage
                 },
-                coord_lexp.handle,
+                coord_handle,
                 coord_type_handle,
-                type_arena,
-                expressions,
+                ctx.type_arena,
+                ctx.expressions,
             ),
             _ => return Err(Error::InvalidImage(image_ty)),
         };
@@ -399,8 +402,9 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         self.lookup_expression.insert(
             result_id,
             LookupExpression {
-                handle: expressions.append(expr, self.span_from_with_op(start)),
+                handle: ctx.expressions.append(expr, self.span_from_with_op(start)),
                 type_id: result_type_id,
+                block_id,
             },
         );
         Ok(())
@@ -411,11 +415,11 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         &mut self,
         mut words_left: u16,
         options: SamplingOptions,
-        type_arena: &Arena<crate::Type>,
-        global_arena: &Arena<crate::GlobalVariable>,
-        arguments: &[FunctionArgument],
-        expressions: &mut Arena<crate::Expression>,
-        parameters_sampling: &mut [SamplingFlags],
+        ctx: &mut super::BlockContext,
+        emitter: &mut crate::front::Emitter,
+        block: &mut crate::Block,
+        block_id: spirv::Word,
+        body_idx: usize,
     ) -> Result<(), Error> {
         let start = self.data_offset;
         let result_type_id = self.next()?;
@@ -442,13 +446,17 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             match spirv::ImageOperands::from_bits_truncate(bit) {
                 spirv::ImageOperands::BIAS => {
                     let bias_expr = self.next()?;
-                    let bias_handle = self.lookup_expression.lookup(bias_expr)?.handle;
+                    let bias_lexp = self.lookup_expression.lookup(bias_expr)?;
+                    let bias_handle =
+                        self.get_expr_handle(bias_expr, bias_lexp, ctx, emitter, block, body_idx);
                     level = crate::SampleLevel::Bias(bias_handle);
                     words_left -= 1;
                 }
                 spirv::ImageOperands::LOD => {
                     let lod_expr = self.next()?;
-                    let lod_handle = self.lookup_expression.lookup(lod_expr)?.handle;
+                    let lod_lexp = self.lookup_expression.lookup(lod_expr)?;
+                    let lod_handle =
+                        self.get_expr_handle(lod_expr, lod_lexp, ctx, emitter, block, body_idx);
                     level = if options.compare {
                         log::debug!("Assuming {:?} is zero", lod_handle);
                         crate::SampleLevel::Zero
@@ -459,9 +467,25 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
                 }
                 spirv::ImageOperands::GRAD => {
                     let grad_x_expr = self.next()?;
-                    let grad_x_handle = self.lookup_expression.lookup(grad_x_expr)?.handle;
+                    let grad_x_lexp = self.lookup_expression.lookup(grad_x_expr)?;
+                    let grad_x_handle = self.get_expr_handle(
+                        grad_x_expr,
+                        grad_x_lexp,
+                        ctx,
+                        emitter,
+                        block,
+                        body_idx,
+                    );
                     let grad_y_expr = self.next()?;
-                    let grad_y_handle = self.lookup_expression.lookup(grad_y_expr)?.handle;
+                    let grad_y_lexp = self.lookup_expression.lookup(grad_y_expr)?;
+                    let grad_y_handle = self.get_expr_handle(
+                        grad_y_expr,
+                        grad_y_lexp,
+                        ctx,
+                        emitter,
+                        block,
+                        body_idx,
+                    );
                     level = if options.compare {
                         log::debug!(
                             "Assuming gradients {:?} and {:?} are not greater than 1",
@@ -496,6 +520,8 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
 
         let si_lexp = self.lookup_sampled_image.lookup(sampled_image_id)?;
         let coord_lexp = self.lookup_expression.lookup(coordinate_id)?;
+        let coord_handle =
+            self.get_expr_handle(coordinate_id, coord_lexp, ctx, emitter, block, body_idx);
         let coord_type_handle = self.lookup_type.lookup(coord_lexp.type_id)?.handle;
 
         let sampling_bit = if options.compare {
@@ -504,48 +530,78 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             SamplingFlags::REGULAR
         };
 
-        let image_ty = match expressions[si_lexp.image] {
+        let image_ty = match ctx.expressions[si_lexp.image] {
             crate::Expression::GlobalVariable(handle) => {
                 if let Some(flags) = self.handle_sampling.get_mut(&handle) {
                     *flags |= sampling_bit;
                 }
 
-                global_arena[handle].ty
+                ctx.global_arena[handle].ty
             }
             crate::Expression::FunctionArgument(i) => {
-                parameters_sampling[i as usize] |= sampling_bit;
-                arguments[i as usize].ty
+                ctx.parameter_sampling[i as usize] |= sampling_bit;
+                ctx.arguments[i as usize].ty
             }
             ref other => return Err(Error::InvalidGlobalVar(other.clone())),
         };
-        match expressions[si_lexp.sampler] {
+        match ctx.expressions[si_lexp.sampler] {
             crate::Expression::GlobalVariable(handle) => {
                 *self.handle_sampling.get_mut(&handle).unwrap() |= sampling_bit
             }
             crate::Expression::FunctionArgument(i) => {
-                parameters_sampling[i as usize] |= sampling_bit;
+                ctx.parameter_sampling[i as usize] |= sampling_bit;
             }
             ref other => return Err(Error::InvalidGlobalVar(other.clone())),
         }
 
-        let (coordinate, array_index) = match type_arena[image_ty].inner {
+        let ((coordinate, array_index), depth_ref) = match ctx.type_arena[image_ty].inner {
             crate::TypeInner::Image {
                 dim,
                 arrayed,
                 class: _,
-            } => extract_image_coordinates(
-                dim,
-                if options.project {
-                    ExtraCoordinate::Projection
-                } else if arrayed {
-                    ExtraCoordinate::ArrayLayer
-                } else {
-                    ExtraCoordinate::Garbage
+            } => (
+                extract_image_coordinates(
+                    dim,
+                    if options.project {
+                        ExtraCoordinate::Projection
+                    } else if arrayed {
+                        ExtraCoordinate::ArrayLayer
+                    } else {
+                        ExtraCoordinate::Garbage
+                    },
+                    coord_handle,
+                    coord_type_handle,
+                    ctx.type_arena,
+                    ctx.expressions,
+                ),
+                {
+                    match dref_id {
+                        Some(id) => {
+                            let mut expr = self.lookup_expression.lookup(id)?.handle;
+
+                            if options.project {
+                                let required_size = dim.required_coordinate_size();
+                                let right = ctx.expressions.append(
+                                    crate::Expression::AccessIndex {
+                                        base: coord_handle,
+                                        index: required_size.map_or(1, |size| size as u32),
+                                    },
+                                    crate::Span::default(),
+                                );
+                                expr = ctx.expressions.append(
+                                    crate::Expression::Binary {
+                                        op: crate::BinaryOperator::Divide,
+                                        left: expr,
+                                        right,
+                                    },
+                                    crate::Span::default(),
+                                )
+                            };
+                            Some(expr)
+                        }
+                        None => None,
+                    }
                 },
-                coord_lexp.handle,
-                coord_type_handle,
-                type_arena,
-                expressions,
             ),
             _ => return Err(Error::InvalidImage(image_ty)),
         };
@@ -557,16 +613,14 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             array_index,
             offset,
             level,
-            depth_ref: match dref_id {
-                Some(id) => Some(self.lookup_expression.lookup(id)?.handle),
-                None => None,
-            },
+            depth_ref,
         };
         self.lookup_expression.insert(
             result_id,
             LookupExpression {
-                handle: expressions.append(expr, self.span_from_with_op(start)),
+                handle: ctx.expressions.append(expr, self.span_from_with_op(start)),
                 type_id: result_type_id,
+                block_id,
             },
         );
         Ok(())
@@ -575,7 +629,11 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
     pub(super) fn parse_image_query_size(
         &mut self,
         at_level: bool,
-        expressions: &mut Arena<crate::Expression>,
+        ctx: &mut super::BlockContext,
+        emitter: &mut crate::front::Emitter,
+        block: &mut crate::Block,
+        block_id: spirv::Word,
+        body_idx: usize,
     ) -> Result<(), Error> {
         let start = self.data_offset;
         let result_type_id = self.next()?;
@@ -584,11 +642,13 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         let level = if at_level {
             let level_id = self.next()?;
             let level_lexp = self.lookup_expression.lookup(level_id)?;
-            Some(level_lexp.handle)
+            Some(self.get_expr_handle(level_id, level_lexp, ctx, emitter, block, body_idx))
         } else {
             None
         };
 
+        // No need to call get_expr_handle here since only globals/arguments are
+        // allowed as images and they are always in the root scope
         //TODO: handle arrays and cubes
         let image_lexp = self.lookup_expression.lookup(image_id)?;
 
@@ -599,8 +659,9 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         self.lookup_expression.insert(
             result_id,
             LookupExpression {
-                handle: expressions.append(expr, self.span_from_with_op(start)),
+                handle: ctx.expressions.append(expr, self.span_from_with_op(start)),
                 type_id: result_type_id,
+                block_id,
             },
         );
         Ok(())
@@ -610,12 +671,15 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
         &mut self,
         query: crate::ImageQuery,
         expressions: &mut Arena<crate::Expression>,
+        block_id: spirv::Word,
     ) -> Result<(), Error> {
         let start = self.data_offset;
         let result_type_id = self.next()?;
         let result_id = self.next()?;
         let image_id = self.next()?;
 
+        // No need to call get_expr_handle here since only globals/arguments are
+        // allowed as images and they are always in the root scope
         let image_lexp = self.lookup_expression.lookup(image_id)?.clone();
 
         let expr = crate::Expression::ImageQuery {
@@ -627,6 +691,7 @@ impl<I: Iterator<Item = u32>> super::Parser<I> {
             LookupExpression {
                 handle: expressions.append(expr, self.span_from_with_op(start)),
                 type_id: result_type_id,
+                block_id,
             },
         );
         Ok(())
