@@ -44,6 +44,15 @@ and compound expressions refer to their sub-expressions via `Handle<Expression>`
 values. (When examining the serialized form of a `Module`, note that the first
 element of an `Arena` has an index of 1, not 0.)
 
+A [`UniqueArena`] is just like an `Arena`, except that it stores only a single
+instance of each value. The value type must implement `Eq` and `Hash`. Like an
+`Arena`, inserting a value into a `UniqueArena` returns a `Handle` which can be
+used to efficiently access the value, without a hash lookup. Inserting a value
+multiple times returns the same `Handle`.
+
+If the `span` feature is enabled, both `Arena` and `UniqueArena` can associate a
+source code span with each element.
+
 ## Function Calls
 
 Naga's representation of function calls is unusual. Most languages treat
@@ -53,7 +62,7 @@ returns a value, a call statement designates a particular [`Expression::CallResu
 expression to represent its return value, for use by subsequent statements and
 expressions.
 
-## `Expression` evaluation time and scope
+## `Expression` evaluation time
 
 It is essential to know when an [`Expression`] should be evaluated, because its
 value may depend on previous [`Statement`]s' effects. But whereas the order of
@@ -87,15 +96,51 @@ Naga's rules for when `Expression`s are evaluated are as follows:
         a pointer. Such global variables hold opaque types like shaders or
         images, and cannot be assigned to.
 
--   A [`Call`](Expression::CallResult) expression that is the `result` of a
-    [`Statement::Call`], representing the call's return value, is evaluated when
-    the `Call` statement is executed.
+-   A [`CallResult`] expression that is the `result` of a [`Statement::Call`],
+    representing the call's return value, is evaluated when the `Call` statement
+    is executed.
+
+-   Similarly, an [`AtomicResult`] expression that is the `result` of an
+    [`Atomic`] statement, representing the result of the atomic operation, is
+    evaluated when the `Atomic` statement is executed.
 
 -   All other expressions are evaluated when the (unique) [`Statement::Emit`]
     statement that covers them is executed. The [`Expression::needs_pre_emit`]
     method returns `true` if the given expression is one of those variants that
     does *not* need to be covered by an `Emit` statement.
 
+Now, strictly speaking, not all `Expression` variants actually care when they're
+evaluated. For example, you can evaluate a [`BinaryOperator::Add`] expression
+any time you like, as long as you give it the right operands. It's really only a
+very small set of expressions that are affected by timing:
+
+-   [`Load`], [`ImageSample`], and [`ImageLoad`] expressions are influenced by
+    stores to the variables or images they access, and must execute at the
+    proper time relative to them.
+
+-   [`Derivative`] expressions are sensitive to control flow uniformity: they
+    must not be moved out of an area of uniform control flow into a non-uniform
+    area.
+
+-   More generally, any expression that's used by more than one other expression
+    or statement should probably be evaluated only once, and then stored in a
+    variable to be cited at each point of use.
+
+Naga tries to help back ends handle all these cases correctly in a somewhat
+circuitous way. The [`ModuleInfo`] structure returned by [`Validator::validate`]
+provides a reference count for each expression in each function in the module.
+Naturally, any expression with a reference count of two or more deserves to be
+evaluated and stored in a temporary variable at the point that the `Emit`
+statement covering it is executed. But if we selectively lower the reference
+count threshold to _one_ for the sensitive expression types listed above, so
+that we _always_ generate a temporary variable and save their value, then the
+same code that manages multiply referenced expressions will take care of
+introducing temporaries for time-sensitive expressions as well. The
+`Expression::bake_ref_count` method (private to the back ends) is meant to help
+with this.
+
+## `Expression` scope
+
 Each `Expression` has a *scope*, which is the region of the function within
 which it can be used by `Statement`s and other `Expression`s. It is a validation
 error to use an `Expression` outside its scope.
@@ -110,24 +155,34 @@ An expression's scope is defined as follows:
     subsequent expressions in that `Emit`, the subsequent statements in the `Block`
     to which that `Emit` belongs (if any) and their sub-statements (if any).
 
--   If a [`Call`] statement has a `result` expression, then that expression's
-    scope covers the subsequent statements in the `Block` to which that `Call`
-    belongs (if any) and their sub-statements (if any).
+-   The `result` expression of a [`Call`] or [`Atomic`] statement has a scope
+    covering the subsequent statements in the `Block` in which the statement
+    occurs (if any) and their sub-statements (if any).
 
 For example, this implies that an expression evaluated by some statement in a
 nested `Block` is not available in the `Block`'s parents. Such a value would
 need to be stored in a local variable to be carried upwards in the statement
 tree.
 
-[`Call`]: Statement::Call
+[`AtomicResult`]: Expression::AtomicResult
+[`CallResult`]: Expression::CallResult
 [`Constant`]: Expression::Constant
-[`Emit`]: Statement::Emit
+[`Derivative`]: Expression::Derivative
 [`FunctionArgument`]: Expression::FunctionArgument
 [`GlobalVariable`]: Expression::GlobalVariable
+[`ImageLoad`]: Expression::ImageLoad
+[`ImageSample`]: Expression::ImageSample
 [`Load`]: Expression::Load
 [`LocalVariable`]: Expression::LocalVariable
+
+[`Atomic`]: Statement::Atomic
+[`Call`]: Statement::Call
+[`Emit`]: Statement::Emit
 [`Store`]: Statement::Store
 
+[`Validator::validate`]: valid::Validator::validate
+[`ModuleInfo`]: valid::ModuleInfo
+
 !*/
 
 // TODO: use `strip_prefix` instead when Rust 1.45 <= MSRV
@@ -157,7 +212,7 @@ pub mod proc;
 mod span;
 pub mod valid;
 
-pub use crate::arena::{Arena, Handle, Range};
+pub use crate::arena::{Arena, Handle, Range, UniqueArena};
 
 use std::{
     collections::{HashMap, HashSet},
@@ -261,6 +316,7 @@ pub enum StorageClass {
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub enum BuiltIn {
     Position,
+    ViewIndex,
     // vertex
     BaseInstance,
     BaseVertex,
@@ -365,7 +421,7 @@ pub enum Sampling {
 
 /// Member of a user-defined structure.
 // Clone is used only for error reporting and is not intended for end users
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, Debug, Eq, Hash, PartialEq)]
 #[cfg_attr(feature = "serialize", derive(Serialize))]
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub struct StructMember {
@@ -484,7 +540,7 @@ pub enum ImageClass {
 }
 
 /// A data type declared in the module.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Eq, Hash, PartialEq)]
 #[cfg_attr(feature = "serialize", derive(Serialize))]
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub struct Type {
@@ -495,7 +551,7 @@ pub struct Type {
 }
 
 /// Enum with additional information, depending on the kind of type.
-#[derive(Debug, PartialEq)]
+#[derive(Debug, Eq, Hash, PartialEq)]
 #[cfg_attr(feature = "serialize", derive(Serialize))]
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub enum TypeInner {
@@ -517,6 +573,10 @@ pub enum TypeInner {
     Atomic { kind: ScalarKind, width: Bytes },
     /// Pointer to another type.
     ///
+    /// Pointers to scalars and vectors should be treated as equivalent to
+    /// [`ValuePointer`] types. Use the [`TypeInner::equivalent`] method to
+    /// compare types in a way that treats pointers correctly.
+    ///
     /// ## Pointers to non-`SIZED` types
     ///
     /// The `base` type of a pointer may be a non-[`SIZED`] type like a
@@ -534,13 +594,26 @@ pub enum TypeInner {
     /// [`DATA`]: valid::TypeFlags::DATA
     /// [`Array`]: TypeInner::Array
     /// [`Struct`]: TypeInner::Struct
+    /// [`ValuePointer`]: TypeInner::ValuePointer
     /// [`GlobalVariable`]: Expression::GlobalVariable
     /// [`AccessIndex`]: Expression::AccessIndex
     Pointer {
         base: Handle<Type>,
         class: StorageClass,
     },
-    /// Pointer to a value.
+
+    /// Pointer to a scalar or vector.
+    ///
+    /// A `ValuePointer` type is equivalent to a `Pointer` whose `base` is a
+    /// `Scalar` or `Vector` type. This is for use in [`TypeResolution::Value`]
+    /// variants; see the documentation for [`TypeResolution`] for details.
+    ///
+    /// Use the [`TypeInner::equivalent`] method to compare types that could be
+    /// pointers, to ensure that `Pointer` and `ValuePointer` types are
+    /// recognized as equivalent.
+    ///
+    /// [`TypeResolution`]: proc::TypeResolution
+    /// [`TypeResolution::Value`]: proc::TypeResolution::Value
     ValuePointer {
         size: Option<VectorSize>,
         kind: ScalarKind,
@@ -651,13 +724,29 @@ pub enum ConstantInner {
 }
 
 /// Describes how an input/output variable is to be bound.
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, Debug, Eq, PartialEq, Hash)]
 #[cfg_attr(feature = "serialize", derive(Serialize))]
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub enum Binding {
     /// Built-in shader variable.
     BuiltIn(BuiltIn),
+
     /// Indexed location.
+    ///
+    /// Values passed from the [`Vertex`] stage to the [`Fragment`] stage must
+    /// have their `interpolation` defaulted (i.e. not `None`) by the front end
+    /// as appropriate for that language.
+    ///
+    /// For other stages, we permit interpolations even though they're ignored.
+    /// When a front end is parsing a struct type, it usually doesn't know what
+    /// stages will be using it for IO, so it's easiest if it can apply the
+    /// defaults to anything with a `Location` binding, just in case.
+    ///
+    /// For anything other than floating-point scalars and vectors, the
+    /// interpolation must be `Flat`.
+    ///
+    /// [`Vertex`]: crate::ShaderStage::Vertex
+    /// [`Fragment`]: crate::ShaderStage::Fragment
     Location {
         location: u32,
         interpolation: Option<Interpolation>,
@@ -724,6 +813,7 @@ pub enum BinaryOperator {
     Subtract,
     Multiply,
     Divide,
+    /// Equivalent of the WGSL's `%` operator or SPIR-V's `OpFRem`
     Modulo,
     Equal,
     NotEqual,
@@ -750,6 +840,7 @@ pub enum BinaryOperator {
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub enum AtomicFunction {
     Add,
+    Subtract,
     And,
     ExclusiveOr,
     InclusiveOr,
@@ -802,6 +893,9 @@ pub enum MathFunction {
     Asin,
     Atan,
     Atan2,
+    Asinh,
+    Acosh,
+    Atanh,
     // decomposition
     Ceil,
     Floor,
@@ -841,6 +935,20 @@ pub enum MathFunction {
     // bits
     CountOneBits,
     ReverseBits,
+    ExtractBits,
+    InsertBits,
+    // data packing
+    Pack4x8snorm,
+    Pack4x8unorm,
+    Pack2x16snorm,
+    Pack2x16unorm,
+    Pack2x16float,
+    // data unpacking
+    Unpack4x8snorm,
+    Unpack4x8unorm,
+    Unpack2x16snorm,
+    Unpack2x16unorm,
+    Unpack2x16float,
 }
 
 /// Sampling modifier to control the level of detail.
@@ -926,17 +1034,12 @@ pub enum Expression {
     /// Indexing a [`Vector`] or [`Array`] produces a value of its element type.
     /// Indexing a [`Matrix`] produces a [`Vector`].
     ///
-    /// Indexing a [`Pointer`] to an [`Array`] produces a [`Pointer`] to its
-    /// `base` type, taking on the `Pointer`'s storage class.
-    ///
-    /// Indexing a [`Pointer`] to a [`Vector`] produces a [`ValuePointer`] whose
-    /// size is `None`, taking on the [`Vector`]'s scalar kind and width and the
-    /// [`Pointer`]'s storage class.
-    ///
-    /// Indexing a [`Pointer`] to a [`Matrix`] produces a [`ValuePointer`] for a
-    /// column of the matrix: its size is the matrix's height, its `kind` is
-    /// [`Float`], and it inherits the [`Matrix`]'s width and the [`Pointer`]'s
-    /// storage class.
+    /// Indexing a [`Pointer`] to any of the above produces a pointer to the
+    /// element/component type, in the same [`class`]. In the case of [`Array`],
+    /// the result is an actual [`Pointer`], but for vectors and matrices, there
+    /// may not be any type in the arena representing the component's type, so
+    /// those produce [`ValuePointer`] types equivalent to the appropriate
+    /// [`Pointer`].
     ///
     /// ## Dynamic indexing restrictions
     ///
@@ -961,6 +1064,7 @@ pub enum Expression {
     /// [`Matrix`]: TypeInner::Matrix
     /// [`Array`]: TypeInner::Array
     /// [`Pointer`]: TypeInner::Pointer
+    /// [`class`]: TypeInner::Pointer::class
     /// [`ValuePointer`]: TypeInner::ValuePointer
     /// [`Float`]: ScalarKind::Float
     Access {
@@ -1041,7 +1145,22 @@ pub enum Expression {
         level: SampleLevel,
         depth_ref: Option<Handle<Expression>>,
     },
+
     /// Load a texel from an image.
+    ///
+    /// For most images, this returns a four-element vector of the same
+    /// [`ScalarKind`] as the image. If the format of the image does not have
+    /// four components, default values are provided: the first three components
+    /// (typically R, G, and B) default to zero, and the final component
+    /// (typically alpha) defaults to one.
+    ///
+    /// However, if the image's [`class`] is [`Depth`], then this returns a
+    /// [`Float`] scalar value.
+    ///
+    /// [`ScalarKind`]: ScalarKind
+    /// [`class`]: TypeInner::Image::class
+    /// [`Depth`]: ImageClass::Depth
+    /// [`Float`]: ScalarKind::Float
     ImageLoad {
         /// The image to load a texel from. This must have type [`Image`]. (This
         /// will necessarily be a [`GlobalVariable`] or [`FunctionArgument`]
@@ -1082,11 +1201,9 @@ pub enum Expression {
         /// -   [`Storage`] images hold exactly one sample per texel, so `index` must
         ///     be `None`.
         ///
-        /// -   [`Depth`] images may have mipmaps, so `index` must be `Some(level)`,
-        ///     where `level` identifies the level of detail.
-        ///
-        /// -   [`Sampled`] images may be multisampled or have mipmaps, but not both.
-        ///     Which one is indicated by the `Sampled` variant's [`multi`] field:
+        /// -   [`Depth`] and [`Sampled`] images may be multisampled or have
+        ///     mipmaps, but not both. Which one is indicated by the variant's
+        ///     [`multi`] field:
         ///
         ///     - If `multi` is `true`, then the image has multiple samples per
         ///       texel, and `index` must be `Some(sample)`, where `sample` is
@@ -1109,6 +1226,7 @@ pub enum Expression {
         /// [`multi`]: ImageClass::Sampled::multi
         index: Option<Handle<Expression>>,
     },
+
     /// Query information from an image.
     ImageQuery {
         image: Handle<Expression>,
@@ -1152,6 +1270,7 @@ pub enum Expression {
         arg: Handle<Expression>,
         arg1: Option<Handle<Expression>>,
         arg2: Option<Handle<Expression>>,
+        arg3: Option<Handle<Expression>>,
     },
     /// Cast a simple type to another kind.
     As {
@@ -1189,7 +1308,7 @@ pub use block::Block;
 pub struct SwitchCase {
     /// Value, upon which the case is considered true.
     pub value: i32,
-    /// Body of the cae.
+    /// Body of the case.
     pub body: Block,
     /// If true, the control flow continues to the next case in the list,
     /// or default.
@@ -1207,7 +1326,7 @@ pub enum Statement {
     ///
     /// See the [module-level documentation][emit] for details.
     ///
-    /// [emit]: index.html#expression-evaluation-time-and-scope
+    /// [emit]: index.html#expression-evaluation-time
     Emit(Range<Expression>),
     /// A block containing more statements, to be executed sequentially.
     Block(Block),
@@ -1325,7 +1444,9 @@ pub enum Statement {
         fun: AtomicFunction,
         /// Value to use in the function.
         value: Handle<Expression>,
-        /// Emitted expression as a result.
+        /// [`AtomicResult`] expression representing this function's result.
+        ///
+        /// [`AtomicResult`]: crate::Expression::AtomicResult
         result: Handle<Expression>,
     },
     /// Calls a function.
@@ -1379,6 +1500,9 @@ pub struct Function {
     /// Local variables defined and used in the function.
     pub local_variables: Arena<LocalVariable>,
     /// Expressions used inside this function.
+    ///
+    /// An `Expression` must occur before all other `Expression`s that use its
+    /// value.
     pub expressions: Arena<Expression>,
     /// Map of expressions that have associated variable names
     pub named_expressions: NamedExpressions,
@@ -1431,6 +1555,8 @@ pub struct Function {
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub struct EntryPoint {
     /// Name of this entry point, visible externally.
+    ///
+    /// Entry point names for a given `stage` must be distinct within a module.
     pub name: String,
     /// Shader stage.
     pub stage: ShaderStage,
@@ -1458,12 +1584,15 @@ pub struct EntryPoint {
 #[cfg_attr(feature = "deserialize", derive(Deserialize))]
 pub struct Module {
     /// Storage for the types defined in this module.
-    pub types: Arena<Type>,
+    pub types: UniqueArena<Type>,
     /// Storage for the constants defined in this module.
     pub constants: Arena<Constant>,
     /// Storage for the global variables defined in this module.
     pub global_variables: Arena<GlobalVariable>,
     /// Storage for the functions defined in this module.
+    ///
+    /// Each function must appear in this arena strictly before all its callers.
+    /// Recursion is not supported.
     pub functions: Arena<Function>,
     /// Entry points.
     pub entry_points: Vec<EntryPoint>,
