@@ -21,18 +21,23 @@ pub enum NameKey {
 /// that may need identifiers in a textual backend.
 #[derive(Default)]
 pub struct Namer {
-    unique: FastHashMap<(String, u32), u32>,
+    /// The last numeric suffix used for each base name.
+    unique: FastHashMap<String, u32>,
     keywords: FastHashSet<String>,
-    /// Currently active namespace.
-    namespace_index: u32,
     reserved_prefixes: Vec<String>,
 }
 
 impl Namer {
+    /// Return a form of `string` suitable for use as the base of an identifier.
+    ///
+    /// Retain only alphanumeric and `_` characters. Drop leading digits. Ensure
+    /// that the string does not end with a digit, so we can attach numeric
+    /// suffixes without merging. Avoid prefixes in
+    /// [`Namer::reserved_prefixes`].
     fn sanitize(&self, string: &str) -> String {
         let mut base = string
             .chars()
-            .skip_while(|c| c.is_numeric())
+            .skip_while(|&c| c.is_numeric() || c == '_')
             .filter(|&c| c.is_ascii_alphanumeric() || c == '_')
             .collect::<String>();
         // close the name by '_' if the re is a number, so that
@@ -51,38 +56,25 @@ impl Namer {
         base
     }
 
-    /// Helper function that return unique name without cache update.
-    /// This function should be used **after** [`Namer`](crate::proc::Namer) initialization by [`reset`](Self::reset()) function.
-    pub fn call_unique(&mut self, string: &str) -> String {
-        let base = self.sanitize(string);
-        match self.unique.entry((base, self.namespace_index)) {
-            Entry::Occupied(mut e) => {
-                *e.get_mut() += 1;
-                format!("{}{}", e.key().0, e.get())
-            }
-            Entry::Vacant(e) => {
-                let name = &e.key().0;
-                if self.keywords.contains(&e.key().0) {
-                    let name = format!("{}1", name);
-                    e.insert(1);
-                    name
-                } else {
-                    name.to_string()
-                }
-            }
-        }
-    }
-
+    /// Return a new identifier based on `label_raw`.
+    ///
+    /// The result:
+    /// - is a valid identifier even if `label_raw` is not
+    /// - conflicts with no keywords listed in `Namer::keywords`, and
+    /// - is different from any identifier previously constructed by this
+    ///   `Namer`.
+    ///
+    /// Guarantee uniqueness by applying a numeric suffix when necessary.
     pub fn call(&mut self, label_raw: &str) -> String {
         let base = self.sanitize(label_raw);
-        match self.unique.entry((base, self.namespace_index)) {
+        match self.unique.entry(base) {
             Entry::Occupied(mut e) => {
                 *e.get_mut() += 1;
-                format!("{}{}", e.key().0, e.get())
+                format!("{}{}", e.key(), e.get())
             }
             Entry::Vacant(e) => {
-                let name = &e.key().0;
-                if self.keywords.contains(&e.key().0) {
+                let name = e.key();
+                if self.keywords.contains(e.key()) {
                     let name = format!("{}1", name);
                     e.insert(1);
                     name
@@ -95,19 +87,23 @@ impl Namer {
         }
     }
 
-    fn call_or(&mut self, label: &Option<String>, fallback: &str) -> String {
+    pub fn call_or(&mut self, label: &Option<String>, fallback: &str) -> String {
         self.call(match *label {
             Some(ref name) => name,
             None => fallback,
         })
     }
 
-    fn namespace(&mut self, f: impl FnOnce(&mut Self)) {
-        self.namespace_index += 1;
-        f(self);
-        let current_ns = self.namespace_index;
-        self.unique.retain(|&(_, ns), _| ns != current_ns);
-        self.namespace_index -= 1;
+    /// Enter a local namespace for things like structs.
+    ///
+    /// Struct member names only need to be unique amongst themselves, not
+    /// globally. This function temporarily establishes a fresh, empty naming
+    /// context for the duration of the call to `body`.
+    fn namespace(&mut self, capacity: usize, body: impl FnOnce(&mut Self)) {
+        let fresh = FastHashMap::with_capacity_and_hasher(capacity, Default::default());
+        let outer = std::mem::replace(&mut self.unique, fresh);
+        body(self);
+        self.unique = outer;
     }
 
     pub fn reset(
@@ -133,7 +129,7 @@ impl Namer {
 
             if let crate::TypeInner::Struct { ref members, .. } = ty.inner {
                 // struct members have their own namespace, because access is always prefixed
-                self.namespace(|namer| {
+                self.namespace(members.len(), |namer| {
                     for (index, member) in members.iter().enumerate() {
                         let name = namer.call_or(&member.name, "member");
                         output.insert(NameKey::StructMember(ty_handle, index as u32), name);
