@@ -1,6 +1,10 @@
-use super::{compose::validate_compose, ComposeError, FunctionInfo, ShaderStages, TypeFlags};
+#[cfg(feature = "validate")]
+use super::{compose::validate_compose, FunctionInfo, ShaderStages, TypeFlags};
+#[cfg(feature = "validate")]
+use crate::arena::UniqueArena;
+
 use crate::{
-    arena::{Arena, Handle},
+    arena::Handle,
     proc::{ProcError, ResolveError},
 };
 
@@ -40,7 +44,7 @@ pub enum ExpressionError {
     #[error("Swizzle component {0:?} is outside of vector size {1:?}")]
     InvalidSwizzleComponent(crate::SwizzleComponent, crate::VectorSize),
     #[error(transparent)]
-    Compose(#[from] ComposeError),
+    Compose(#[from] super::ComposeError),
     #[error(transparent)]
     Proc(#[from] ProcError),
     #[error("Operation {0:?} can't work with {1:?}")]
@@ -111,12 +115,14 @@ pub enum ExpressionError {
     InvalidAtomicResultType(crate::ScalarKind, crate::Bytes),
 }
 
+#[cfg(feature = "validate")]
 struct ExpressionTypeResolver<'a> {
     root: Handle<crate::Expression>,
-    types: &'a Arena<crate::Type>,
+    types: &'a UniqueArena<crate::Type>,
     info: &'a FunctionInfo,
 }
 
+#[cfg(feature = "validate")]
 impl<'a> ExpressionTypeResolver<'a> {
     fn resolve(
         &self,
@@ -130,6 +136,7 @@ impl<'a> ExpressionTypeResolver<'a> {
     }
 }
 
+#[cfg(feature = "validate")]
 impl super::Validator {
     pub(super) fn validate_expression(
         &self,
@@ -795,7 +802,14 @@ impl super::Validator {
                     Ti::Scalar {
                         kind: Sk::Bool,
                         width: _,
-                    } => accept_inner.is_sized(),
+                    } => {
+                        // When `condition` is a single boolean, `accept` and
+                        // `reject` can be vectors or scalars.
+                        match *accept_inner {
+                            Ti::Scalar { .. } | Ti::Vector { .. } => true,
+                            _ => false,
+                        }
+                    }
                     Ti::Vector {
                         size,
                         kind: Sk::Bool,
@@ -856,15 +870,17 @@ impl super::Validator {
                 arg,
                 arg1,
                 arg2,
+                arg3,
             } => {
                 use crate::MathFunction as Mf;
 
                 let arg_ty = resolver.resolve(arg)?;
                 let arg1_ty = arg1.map(|expr| resolver.resolve(expr)).transpose()?;
                 let arg2_ty = arg2.map(|expr| resolver.resolve(expr)).transpose()?;
+                let arg3_ty = arg3.map(|expr| resolver.resolve(expr)).transpose()?;
                 match fun {
                     Mf::Abs => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         let good = match *arg_ty {
@@ -876,8 +892,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Min | Mf::Max => {
-                        let arg1_ty = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), None) => ty1,
+                        let arg1_ty = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), None, None) => ty1,
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         let good = match *arg_ty {
@@ -896,8 +912,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Clamp => {
-                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), Some(ty2)) => (ty1, ty2),
+                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), None) => (ty1, ty2),
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         let good = match *arg_ty {
@@ -931,6 +947,9 @@ impl super::Validator {
                     | Mf::Acos
                     | Mf::Asin
                     | Mf::Atan
+                    | Mf::Asinh
+                    | Mf::Acosh
+                    | Mf::Atanh
                     | Mf::Ceil
                     | Mf::Floor
                     | Mf::Round
@@ -944,7 +963,7 @@ impl super::Validator {
                     | Mf::Sign
                     | Mf::Sqrt
                     | Mf::InverseSqrt => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         match *arg_ty {
@@ -958,8 +977,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Atan2 | Mf::Pow | Mf::Distance | Mf::Step => {
-                        let arg1_ty = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), None) => ty1,
+                        let arg1_ty = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), None, None) => ty1,
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         match *arg_ty {
@@ -980,8 +999,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Modf | Mf::Frexp | Mf::Ldexp => {
-                        let arg1_ty = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), None) => ty1,
+                        let arg1_ty = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), None, None) => ty1,
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         let (size0, width0) = match *arg_ty {
@@ -1015,8 +1034,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Dot | Mf::Outer | Mf::Cross | Mf::Reflect => {
-                        let arg1_ty = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), None) => ty1,
+                        let arg1_ty = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), None, None) => ty1,
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         match *arg_ty {
@@ -1034,8 +1053,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Refract => {
-                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), Some(ty2)) => (ty1, ty2),
+                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), None) => (ty1, ty2),
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
 
@@ -1075,7 +1094,7 @@ impl super::Validator {
                         }
                     }
                     Mf::Normalize => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         match *arg_ty {
@@ -1086,8 +1105,8 @@ impl super::Validator {
                         }
                     }
                     Mf::FaceForward | Mf::Fma | Mf::SmoothStep => {
-                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), Some(ty2)) => (ty1, ty2),
+                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), None) => (ty1, ty2),
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         match *arg_ty {
@@ -1115,8 +1134,8 @@ impl super::Validator {
                         }
                     }
                     Mf::Mix => {
-                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty) {
-                            (Some(ty1), Some(ty2)) => (ty1, ty2),
+                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), None) => (ty1, ty2),
                             _ => return Err(ExpressionError::WrongArgumentCount(fun)),
                         };
                         let arg_width = match *arg_ty {
@@ -1155,7 +1174,7 @@ impl super::Validator {
                         }
                     }
                     Mf::Inverse | Mf::Determinant => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         let good = match *arg_ty {
@@ -1167,7 +1186,7 @@ impl super::Validator {
                         }
                     }
                     Mf::Transpose => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         match *arg_ty {
@@ -1176,7 +1195,7 @@ impl super::Validator {
                         }
                     }
                     Mf::CountOneBits | Mf::ReverseBits => {
-                        if arg1_ty.is_some() | arg2_ty.is_some() {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
                             return Err(ExpressionError::WrongArgumentCount(fun));
                         }
                         match *arg_ty {
@@ -1187,6 +1206,118 @@ impl super::Validator {
                             _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
                         }
                     }
+                    Mf::InsertBits => {
+                        let (arg1_ty, arg2_ty, arg3_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), Some(ty3)) => (ty1, ty2, ty3),
+                            _ => return Err(ExpressionError::WrongArgumentCount(fun)),
+                        };
+                        match *arg_ty {
+                            Ti::Scalar { kind: Sk::Sint, .. }
+                            | Ti::Scalar { kind: Sk::Uint, .. }
+                            | Ti::Vector { kind: Sk::Sint, .. }
+                            | Ti::Vector { kind: Sk::Uint, .. } => {}
+                            _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
+                        }
+                        if arg1_ty != arg_ty {
+                            return Err(ExpressionError::InvalidArgumentType(
+                                fun,
+                                1,
+                                arg1.unwrap(),
+                            ));
+                        }
+                        match *arg2_ty {
+                            Ti::Scalar { kind: Sk::Uint, .. } => {}
+                            _ => {
+                                return Err(ExpressionError::InvalidArgumentType(
+                                    fun,
+                                    2,
+                                    arg2.unwrap(),
+                                ))
+                            }
+                        }
+                        match *arg3_ty {
+                            Ti::Scalar { kind: Sk::Uint, .. } => {}
+                            _ => {
+                                return Err(ExpressionError::InvalidArgumentType(
+                                    fun,
+                                    2,
+                                    arg3.unwrap(),
+                                ))
+                            }
+                        }
+                    }
+                    Mf::ExtractBits => {
+                        let (arg1_ty, arg2_ty) = match (arg1_ty, arg2_ty, arg3_ty) {
+                            (Some(ty1), Some(ty2), None) => (ty1, ty2),
+                            _ => return Err(ExpressionError::WrongArgumentCount(fun)),
+                        };
+                        match *arg_ty {
+                            Ti::Scalar { kind: Sk::Sint, .. }
+                            | Ti::Scalar { kind: Sk::Uint, .. }
+                            | Ti::Vector { kind: Sk::Sint, .. }
+                            | Ti::Vector { kind: Sk::Uint, .. } => {}
+                            _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
+                        }
+                        match *arg1_ty {
+                            Ti::Scalar { kind: Sk::Uint, .. } => {}
+                            _ => {
+                                return Err(ExpressionError::InvalidArgumentType(
+                                    fun,
+                                    2,
+                                    arg1.unwrap(),
+                                ))
+                            }
+                        }
+                        match *arg2_ty {
+                            Ti::Scalar { kind: Sk::Uint, .. } => {}
+                            _ => {
+                                return Err(ExpressionError::InvalidArgumentType(
+                                    fun,
+                                    2,
+                                    arg2.unwrap(),
+                                ))
+                            }
+                        }
+                    }
+                    Mf::Pack2x16unorm | Mf::Pack2x16snorm | Mf::Pack2x16float => {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
+                            return Err(ExpressionError::WrongArgumentCount(fun));
+                        }
+                        match *arg_ty {
+                            Ti::Vector {
+                                size: crate::VectorSize::Bi,
+                                kind: Sk::Float,
+                                ..
+                            } => {}
+                            _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
+                        }
+                    }
+                    Mf::Pack4x8snorm | Mf::Pack4x8unorm => {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
+                            return Err(ExpressionError::WrongArgumentCount(fun));
+                        }
+                        match *arg_ty {
+                            Ti::Vector {
+                                size: crate::VectorSize::Quad,
+                                kind: Sk::Float,
+                                ..
+                            } => {}
+                            _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
+                        }
+                    }
+                    Mf::Unpack2x16float
+                    | Mf::Unpack2x16snorm
+                    | Mf::Unpack2x16unorm
+                    | Mf::Unpack4x8snorm
+                    | Mf::Unpack4x8unorm => {
+                        if arg1_ty.is_some() | arg2_ty.is_some() | arg3_ty.is_some() {
+                            return Err(ExpressionError::WrongArgumentCount(fun));
+                        }
+                        match *arg_ty {
+                            Ti::Scalar { kind: Sk::Uint, .. } => {}
+                            _ => return Err(ExpressionError::InvalidArgumentType(fun, 0, arg)),
+                        }
+                    }
                 }
                 ShaderStages::all()
             }
@@ -1221,7 +1352,7 @@ impl super::Validator {
                     if let Some(&Ti::Array {
                         size: crate::ArraySize::Dynamic,
                         ..
-                    }) = resolver.types.try_get(base).map(|ty| &ty.inner)
+                    }) = resolver.types.get_handle(base).map(|ty| &ty.inner)
                     {
                         ShaderStages::all()
                     } else {
