@@ -1,8 +1,13 @@
+use crate::arena::Handle;
+#[cfg(feature = "validate")]
+use crate::arena::{Arena, UniqueArena};
+
 use super::{
     analyzer::{UniformityDisruptor, UniformityRequirements},
-    ExpressionError, FunctionInfo, ModuleInfo, ShaderStages, TypeFlags, ValidationFlags,
+    ExpressionError, FunctionInfo, ModuleInfo,
 };
-use crate::arena::{Arena, Handle};
+
+#[cfg(feature = "validate")]
 use bit_set::BitSet;
 
 #[derive(Clone, Debug, thiserror::Error)]
@@ -76,6 +81,12 @@ pub enum FunctionError {
     },
     #[error("Argument '{name}' at index {index} has a type that can't be passed into functions.")]
     InvalidArgumentType { index: usize, name: String },
+    #[error("Argument '{name}' at index {index} is a pointer of class {class:?}, which can't be passed into functions.")]
+    InvalidArgumentPointerClass {
+        index: usize,
+        name: String,
+        class: crate::StorageClass,
+    },
     #[error("There are instructions after `return`/`break`/`continue`")]
     InstructionsAfterReturn,
     #[error("The `break` is used outside of a `loop` or `switch` context")]
@@ -90,7 +101,7 @@ pub enum FunctionError {
     InvalidIfType(Handle<crate::Expression>),
     #[error("The `switch` value {0:?} is not an integer scalar")]
     InvalidSwitchType(Handle<crate::Expression>),
-    #[error("Multiple `switch` cases for {0} are present")]
+    #[error("Multiple `switch` cases for {0:?} are present")]
     ConflictingSwitchCase(i32),
     #[error("The pointer {0:?} doesn't relate to a valid destination for a store")]
     InvalidStorePointer(Handle<crate::Expression>),
@@ -135,17 +146,25 @@ bitflags::bitflags! {
     }
 }
 
+#[cfg(feature = "validate")]
+struct BlockInfo {
+    stages: super::ShaderStages,
+    finished: bool,
+}
+
+#[cfg(feature = "validate")]
 struct BlockContext<'a> {
     abilities: ControlFlowAbility,
     info: &'a FunctionInfo,
     expressions: &'a Arena<crate::Expression>,
-    types: &'a Arena<crate::Type>,
+    types: &'a UniqueArena<crate::Type>,
     global_vars: &'a Arena<crate::GlobalVariable>,
     functions: &'a Arena<crate::Function>,
     prev_infos: &'a [FunctionInfo],
     return_type: Option<Handle<crate::Type>>,
 }
 
+#[cfg(feature = "validate")]
 impl<'a> BlockContext<'a> {
     fn new(
         fun: &'a crate::Function,
@@ -217,13 +236,14 @@ impl<'a> BlockContext<'a> {
 }
 
 impl super::Validator {
+    #[cfg(feature = "validate")]
     fn validate_call(
         &mut self,
         function: Handle<crate::Function>,
         arguments: &[Handle<crate::Expression>],
         result: Option<Handle<crate::Expression>>,
         context: &BlockContext,
-    ) -> Result<ShaderStages, CallError> {
+    ) -> Result<super::ShaderStages, CallError> {
         let fun = context
             .functions
             .try_get(function)
@@ -238,7 +258,8 @@ impl super::Validator {
             let ty = context
                 .resolve_type_impl(expr, &self.valid_expression_set)
                 .map_err(|error| CallError::Argument { index, error })?;
-            if ty != &context.types[arg.ty].inner {
+            let arg_inner = &context.types[arg.ty].inner;
+            if !ty.equivalent(arg_inner, context.types) {
                 return Err(CallError::ArgumentType {
                     index,
                     required: arg.ty,
@@ -266,6 +287,7 @@ impl super::Validator {
         Ok(callee_info.available_stages)
     }
 
+    #[cfg(feature = "validate")]
     fn validate_atomic(
         &mut self,
         pointer: Handle<crate::Expression>,
@@ -322,14 +344,15 @@ impl super::Validator {
         Ok(())
     }
 
+    #[cfg(feature = "validate")]
     fn validate_block_impl(
         &mut self,
         statements: &[crate::Statement],
         context: &BlockContext,
-    ) -> Result<ShaderStages, FunctionError> {
+    ) -> Result<BlockInfo, FunctionError> {
         use crate::{Statement as S, TypeInner as Ti};
         let mut finished = false;
-        let mut stages = ShaderStages::all();
+        let mut stages = super::ShaderStages::all();
         for statement in statements {
             if finished {
                 return Err(FunctionError::InstructionsAfterReturn);
@@ -345,7 +368,9 @@ impl super::Validator {
                     }
                 }
                 S::Block(ref block) => {
-                    stages &= self.validate_block(block, context)?;
+                    let info = self.validate_block(block, context)?;
+                    stages &= info.stages;
+                    finished = info.finished;
                 }
                 S::If {
                     condition,
@@ -359,8 +384,8 @@ impl super::Validator {
                         } => {}
                         _ => return Err(FunctionError::InvalidIfType(condition)),
                     }
-                    stages &= self.validate_block(accept, context)?;
-                    stages &= self.validate_block(reject, context)?;
+                    stages &= self.validate_block(accept, context)?.stages;
+                    stages &= self.validate_block(reject, context)?.stages;
                 }
                 S::Switch {
                     selector,
@@ -368,6 +393,10 @@ impl super::Validator {
                     ref default,
                 } => {
                     match *context.resolve_type(selector, &self.valid_expression_set)? {
+                        Ti::Scalar {
+                            kind: crate::ScalarKind::Uint,
+                            width: _,
+                        } => {}
                         Ti::Scalar {
                             kind: crate::ScalarKind::Sint,
                             width: _,
@@ -385,9 +414,9 @@ impl super::Validator {
                     let sub_context =
                         context.with_abilities(pass_through_abilities | ControlFlowAbility::BREAK);
                     for case in cases {
-                        stages &= self.validate_block(&case.body, &sub_context)?;
+                        stages &= self.validate_block(&case.body, &sub_context)?.stages;
                     }
-                    stages &= self.validate_block(default, &sub_context)?;
+                    stages &= self.validate_block(default, &sub_context)?.stages;
                 }
                 S::Loop {
                     ref body,
@@ -397,18 +426,22 @@ impl super::Validator {
                     // because the continuing{} block inherits the scope
                     let base_expression_count = self.valid_expression_list.len();
                     let pass_through_abilities = context.abilities & ControlFlowAbility::RETURN;
-                    stages &= self.validate_block_impl(
-                        body,
-                        &context.with_abilities(
-                            pass_through_abilities
-                                | ControlFlowAbility::BREAK
-                                | ControlFlowAbility::CONTINUE,
-                        ),
-                    )?;
-                    stages &= self.validate_block_impl(
-                        continuing,
-                        &context.with_abilities(ControlFlowAbility::empty()),
-                    )?;
+                    stages &= self
+                        .validate_block_impl(
+                            body,
+                            &context.with_abilities(
+                                pass_through_abilities
+                                    | ControlFlowAbility::BREAK
+                                    | ControlFlowAbility::CONTINUE,
+                            ),
+                        )?
+                        .stages;
+                    stages &= self
+                        .validate_block_impl(
+                            continuing,
+                            &context.with_abilities(ControlFlowAbility::empty()),
+                        )?
+                        .stages;
                     for handle in self.valid_expression_list.drain(base_expression_count..) {
                         self.valid_expression_set.remove(handle.index());
                     }
@@ -433,7 +466,17 @@ impl super::Validator {
                         .map(|expr| context.resolve_type(expr, &self.valid_expression_set))
                         .transpose()?;
                     let expected_ty = context.return_type.map(|ty| &context.types[ty].inner);
-                    if value_ty != expected_ty {
+                    // We can't return pointers, but it seems best not to embed that
+                    // assumption here, so use `TypeInner::equivalent` for comparison.
+                    let okay = match (value_ty, expected_ty) {
+                        (None, None) => true,
+                        (Some(value_inner), Some(expected_inner)) => {
+                            value_inner.equivalent(expected_inner, context.types)
+                        }
+                        (_, _) => false,
+                    };
+
+                    if !okay {
                         log::error!(
                             "Returning {:?} where {:?} is expected",
                             value_ty,
@@ -447,7 +490,7 @@ impl super::Validator {
                     finished = true;
                 }
                 S::Barrier(_) => {
-                    stages &= ShaderStages::COMPUTE;
+                    stages &= super::ShaderStages::COMPUTE;
                 }
                 S::Store { pointer, value } => {
                     let mut current = pointer;
@@ -593,32 +636,34 @@ impl super::Validator {
                 }
             }
         }
-        Ok(stages)
+        Ok(BlockInfo { stages, finished })
     }
 
+    #[cfg(feature = "validate")]
     fn validate_block(
         &mut self,
         statements: &[crate::Statement],
         context: &BlockContext,
-    ) -> Result<ShaderStages, FunctionError> {
+    ) -> Result<BlockInfo, FunctionError> {
         let base_expression_count = self.valid_expression_list.len();
-        let stages = self.validate_block_impl(statements, context)?;
+        let info = self.validate_block_impl(statements, context)?;
         for handle in self.valid_expression_list.drain(base_expression_count..) {
             self.valid_expression_set.remove(handle.index());
         }
-        Ok(stages)
+        Ok(info)
     }
 
+    #[cfg(feature = "validate")]
     fn validate_local_var(
         &self,
         var: &crate::LocalVariable,
-        types: &Arena<crate::Type>,
+        types: &UniqueArena<crate::Type>,
         constants: &Arena<crate::Constant>,
     ) -> Result<(), LocalVariableError> {
         log::debug!("var {:?}", var);
         if !self.types[var.ty.index()]
             .flags
-            .contains(TypeFlags::DATA | TypeFlags::SIZED)
+            .contains(super::TypeFlags::DATA | super::TypeFlags::SIZED)
         {
             return Err(LocalVariableError::InvalidType(var.ty));
         }
@@ -649,8 +694,13 @@ impl super::Validator {
         module: &crate::Module,
         mod_info: &ModuleInfo,
     ) -> Result<FunctionInfo, FunctionError> {
+        #[cfg(feature = "validate")]
         let mut info = mod_info.process_function(fun, module, self.flags)?;
 
+        #[cfg(not(feature = "validate"))]
+        let info = mod_info.process_function(fun, module, self.flags)?;
+
+        #[cfg(feature = "validate")]
         for (var_handle, var) in fun.local_variables.iter() {
             self.validate_local_var(var, &module.types, &module.constants)
                 .map_err(|error| FunctionError::LocalVariable {
@@ -660,16 +710,30 @@ impl super::Validator {
                 })?;
         }
 
+        #[cfg(feature = "validate")]
         for (index, argument) in fun.arguments.iter().enumerate() {
             if !self.types[argument.ty.index()]
                 .flags
-                .contains(TypeFlags::ARGUMENT)
+                .contains(super::TypeFlags::ARGUMENT)
             {
                 return Err(FunctionError::InvalidArgumentType {
                     index,
                     name: argument.name.clone().unwrap_or_default(),
                 });
             }
+            match module.types[argument.ty].inner.pointer_class() {
+                Some(crate::StorageClass::Private)
+                | Some(crate::StorageClass::Function)
+                | Some(crate::StorageClass::WorkGroup)
+                | None => {}
+                Some(other) => {
+                    return Err(FunctionError::InvalidArgumentPointerClass {
+                        index,
+                        name: argument.name.clone().unwrap_or_default(),
+                        class: other,
+                    })
+                }
+            }
         }
 
         self.valid_expression_set.clear();
@@ -678,7 +742,8 @@ impl super::Validator {
             if expr.needs_pre_emit() {
                 self.valid_expression_set.insert(handle.index());
             }
-            if self.flags.contains(ValidationFlags::EXPRESSIONS) {
+            #[cfg(feature = "validate")]
+            if self.flags.contains(super::ValidationFlags::EXPRESSIONS) {
                 match self.validate_expression(
                     handle,
                     expr,
@@ -693,11 +758,14 @@ impl super::Validator {
             }
         }
 
-        if self.flags.contains(ValidationFlags::BLOCKS) {
-            let stages = self.validate_block(
-                &fun.body,
-                &BlockContext::new(fun, module, &info, &mod_info.functions),
-            )?;
+        #[cfg(feature = "validate")]
+        if self.flags.contains(super::ValidationFlags::BLOCKS) {
+            let stages = self
+                .validate_block(
+                    &fun.body,
+                    &BlockContext::new(fun, module, &info, &mod_info.functions),
+                )?
+                .stages;
             info.available_stages &= stages;
         }
         Ok(info)
